/*================================================
 * Routine pour creation d'objets 3d
 *------------------------------------------------
 * (c) Disk91 / TeKnO-BeDoUiNs / 1998
 *================================================
 */

 #include <stdio.h>
 #include <allegro.h>
 #include "3d.h"

 int memory_object_f;


/*-------------------------------------------
 * public for DESTROY
 *-------------------------------------------
 */
 void destroy_object_f( OBJECT_f * object )
 {
    if ( object )
    {
       free (object->vtx_begin);
       if ( object->vtx_end ) free (object->vtx_end);
       if ( object->vtx_2d  ) free (object->vtx_2d);

       // puisque les allocations sont regroupees, on se limite a ces free
       free (object->poly_begin[0].index);
       free (object->poly_begin[0].ptvtx3d);

       free (object->poly_begin);
       if ( object->poly_end ) free (object->poly_end);
       if ( object->ptpoly   ) free (object->ptpoly);
       free (object);
    }
 }

/*=================================================
 * Method to create object
 *=================================================
 */

 OBJECT_f * create_object_f( int nb_vertex, int nb_poly, int max_vtx )
 {
    OBJECT_f * ob;
    int base1,base2;
    int i;

    // la taille memoire allouee
    memory_object_f = sizeof(OBJECT_f)
                      + 4 * nb_vertex * sizeof ( V3D_f    )
                      + 2 * nb_poly   * sizeof ( POLY_f   )
                      + 1 * nb_poly   * sizeof ( V3D_f *  ) * (max_vtx)
                      + 2 * nb_poly   * sizeof ( V3D_f *  ) * (max_vtx + 10)
                      + 3 * nb_poly   * sizeof ( int      ) * (max_vtx)
                      + 3 * nb_poly   * sizeof ( int      ) * (max_vtx + 10)
                      +     nb_poly   * sizeof ( POLY_f * ) ;

    // allocation de l'objet
    ob = (OBJECT_f *) malloc( sizeof (OBJECT_f) );
    if ( !ob ) D3ERROR("Allocation d'un objet impossible",40)

    ob->nb_vtx        = nb_vertex;
    ob->nb_poly_begin = nb_poly;

    ob->vtx_begin = ( V3D_f * ) malloc (       nb_vertex        * sizeof ( V3D_f ) );
    ob->vtx_end   = ( V3D_f * ) malloc ( (int)(nb_vertex * 1.5) * sizeof ( V3D_f ) );
    ob->vtx_2d    = ( V3D_f * ) malloc ( (int)(nb_vertex * 1.5) * sizeof ( V3D_f ) );
    if ( !ob->vtx_begin || !ob->vtx_end || !ob->vtx_2d )
       D3ERROR("Allocation des Vertex impossible",41)

    ob->poly_begin = ( POLY_f * ) malloc ( nb_poly * sizeof ( POLY_f ) );
    ob->poly_end   = ( POLY_f * ) malloc ( nb_poly * sizeof ( POLY_f ) );
    if ( !ob->poly_begin || !ob->poly_end )
       D3ERROR("Allocation des Polygones impossible",42)


    // allocation tableaux de pointeurs des faces
    ob->poly_begin[0].index   = ( int *    ) malloc ( nb_poly * (( 6 * max_vtx ) + 30) * sizeof ( int ) );
    ob->poly_begin[0].ptvtx3d = ( V3D_f ** ) malloc ( nb_poly * (( 3 * max_vtx ))      * sizeof ( V3D_f * ) );
    if ( !ob->poly_begin[0].index || !ob->poly_begin[0].ptvtx3d )
    {
       printf(" Memoire desiree : %d\n", memory_object_f);
       D3ERROR("Impossible d'allouer la memoire aux donnees des polygones",43)
    }

    for ( i = 0 ; i < nb_poly ; i++ )
    {
       base1 = i * (( 6*max_vtx ) + 30);
       base2 = i * (( 3*max_vtx )     );
       if ( i != 0 )
       {
          ob->poly_begin[i].ptvtx3d = &ob->poly_begin[0].ptvtx3d[base2];
          ob->poly_end[i].ptvtx2d   = NULL;
          ob->poly_begin[i].index   = &ob->poly_begin[0].index[base1];
       }
       ob->poly_begin[i].u          = &ob->poly_begin[0].index[base1+max_vtx];
       ob->poly_begin[i].v          = &ob->poly_begin[0].index[base1+2*max_vtx];

       ob->poly_end[i].ptvtx3d      = &ob->poly_begin[0].ptvtx3d[base2+max_vtx];
       ob->poly_end[i].ptvtx2d      = &ob->poly_begin[0].ptvtx3d[base2+2*max_vtx];
       ob->poly_end[i].index        = &ob->poly_begin[0].index[base1+3*max_vtx];
       ob->poly_end[i].u            = &ob->poly_begin[0].index[base1+4*max_vtx+10];
       ob->poly_end[i].v            = &ob->poly_begin[0].index[base1+5*max_vtx+20];
    }

    ob->ptpoly     = ( POLY_f ** ) malloc ( nb_poly * sizeof ( POLY_f * ) );
    if ( !ob->ptpoly )
       D3ERROR("Impossible d'allouer de la memoire aux pteurs de polygones",44);


    return ob;
 }

/*---------------------------------------------------------------------------------
 * Reduit la taille d'un objet en suprimant les allocation inutiles pour
 *  les coordonnees du monde uniquement (apres chargement)
 *---------------------------------------------------------------------------------
 */
 void lightify_object_f( OBJECT_f * object )
 {
    if ( object )
    {
       if ( object->vtx_end ) { free (object->vtx_end); object->vtx_end = NULL; }
       if ( object->vtx_2d  ) { free (object->vtx_2d); object->vtx_2d = NULL; }
       if ( object->ptpoly   ) { free (object->ptpoly); object->ptpoly = NULL; }
    }
 }

