/*========================================
 * format d'objet pour mode 3d
 * erreur 4x et 8x pour objet
 *----------------------------------------
 * (c) Disk91 / TeKnO-BeDoUiNs
 *========================================
 */

 #ifndef _3D_H_
 #define _3D_H_
 #include <allegro.h>

 #define D3ERROR(x,y) {  allegro_exit();                         \
                         printf(" Erreur 3D n %d\n  ",y);      \
                         puts(x) ; exit(y);                      \
                       }
 #define OBJECTERROR(x,y) {  allegro_exit();                         \
                             printf(" Erreur OBJECT n %d\n  ",y);  \
                             puts(x) ; exit(y);                      \
                          }


 /*--- nombre maximum de vertex par face ---*/

 /*-------------------------------------
  * Structures
  *-------------------------------------
  */

 typedef struct struct_poly_f
 {

    int        nb_vtx;                   /*--- nombre de vertex dans le poly ---*/
    float      zmoy;                     /*--- zmoyen de la face             ---*/
    int        coul;                     /*--- couleur de la face            ---*/

    float      nx,ny,nz;                 /*--- normale a la face dans espace obj ---*/
    int        state;                    /*--- = 1 si on a un vecteur coherent   ---*/

    int    *   index;                    /*--- index des faces du polygone              ---*/
    V3D_f  **  ptvtx3d;                  /*--- tableau de pointeurs sur les faces 3d    ---*/
    V3D_f  **  ptvtx2d;                  /*--- tableau de pointeurs sur les faces 2d    ---*/

    int        type;                     /*--- type de rendu pour ce polygone           ---*/
    int    *   u;                        /*--- coord u et v pour mapping par point      ---*/
    int    *   v;
    BITMAP *   map;                      /*--- image utilisee pour le mapping           ---*/

                                         /*--- rq : ptvtx2d non utilise dans poly_begin ---*/
                                         /*---  donc non alloue ....                    ---*/

 } POLY_f;

 typedef struct struct_object_f
 {

    int         nb_vtx;                 /*--- nombre total de vertex ---*/
    int         nb_poly_begin;          /*--- nombre de polygons dans l'objet de base ---*/
    int         nb_poly_end;            /*--- nombre de polygons dans l'objet 2d ---*/

    V3D_f   *   vtx_begin;              /*--- tableau des points de base ---*/
    V3D_f   *   vtx_end;                /*--- tableau des points transformes ---*/
    V3D_f   *   vtx_2d;                 /*--- tableau des points projetes ---*/
                                        /*--- remarque, ce dernier est plus grand ---*/
                                        /*--- pour les faces clippes ---*/

    POLY_f  *   poly_begin;             /*--- tableau des faces dans l'objet de base ---*/
    POLY_f  *   poly_end;               /*--- tableau des faces dans l'objet modifie ---*/

    POLY_f  **  ptpoly;                 /*--- pteur sur les faces modifiees pour le trie --*/

 } OBJECT_f;

 /*------------------------------------
  * Prototypes
  *------------------------------------
  */

  extern int memory_object_f;

  OBJECT_f * create_object_f( int nb_vertex, int nb_poly, int max_vtx );

  void       destroy_object_f( OBJECT_f * object );
  void lightify_object_f( OBJECT_f * object );



  void transform_object_f( OBJECT_f * ob, float scale, float xrot, float yrot, float zrot,
                                                      float tx,   float ty,   float tz   );
  void transform4bsp_object_f( OBJECT_f * ob, float scale, float xrot, float yrot, float zrot,
                                                      float tx,   float ty,   float tz   );

  void normalize_object_f( OBJECT_f * ob  );

  void clip3d_object_f( OBJECT_f * ob, int type, float min_z, float max_z );
  void clip3d4bsp_object_f( OBJECT_f * ob, int type, float min_z, float max_z );

  int  clip_f( int type, float min_z, int * nbad, int vc, V3D_f ** in, V3D_f ** out );

#endif                          /*--- _3D_H_ --*/
