/*======================================================================
 * BSP TREE COMPILER a partir du compilateur BSP 2D Zen de la pgm gfx
 *                   + doc bspfaq + pgm BSP 3D par Pin Fei Sun
 *----------------------------------------------------------------------
 * (c) Disk91 /TeKnO-BeDoUiNs/ 12-1998
 *======================================================================
 */

  #include <stdio.h>
  #include <stdlib.h>
  #include <allegro.h>
  #include <math.h>
  #include "3d.h"
  #include "bsp.h"

  #define  BACK    0            // face derriere
  #define  FRONT   1            // face devant
  #define  SPLIT   2            // face coupee
  #define  COPLAN  3            // face coplanaire

  #define  OPTIM   1            // optimisation avec fichier root
  #define  NOOPTIM 0            // pas d'optimisation
  #define  MAXPILE 20000
  #define  MAXROOT 20000

  static int NumCompilePoly = 0;
  static int MaxCompilePoly = 0;
  static int NumCompileVtx  = 0;
  static int MaxCompileVtx  = 0;
  static int FirstAddedVtx  = 0;

  static PACKFILE * rootfile;            // fichier de choix des root
  static int      * rootback;            // file des root
  static int        ptrootback;          // pteur sur file des root
  static int        vtxmerged[MAXPILE];  // les poly auquels on ajoute des faces
  static int        ptvtxmerged;

  static int iter = 0;

extern BITMAP * dblbuff;

/*-----------------------------------------------------
 * Determine ou est un point par rapport a 1 plan
 *-----------------------------------------------------
 */
 static float dist_point( OBJECT_f * obj, POLY_f * plane, V3D_f * x )
 {
    float value;
    float X,Y,Z;

    X = x->x - plane->ptvtx3d[0]->x;
    Y = x->y - plane->ptvtx3d[0]->y;
    Z = x->z - plane->ptvtx3d[0]->z;
    value = dot_product_f(X, Y, Z, plane->nx, plane->ny, plane->nz);

    return value;
 }

/*-----------------------------------------------------
 * Determine ou est la face par rapport a un plan
 *-----------------------------------------------------
 */
 static int classify( OBJECT_f * obj, POLY_f * root, POLY_f * x )
 {
    int   front = 0;
    int   back  = 0;
    int   i;
    float cur;

    for ( i = 0 ; i < x->nb_vtx ; i++ )
    {
      cur = dist_point( obj, root, &obj->vtx_begin[x->index[i]] );
      if (      cur >  0.01 ) front++;
      else if ( cur < -0.01 ) back++;

    }

    if ( (back  > 0) && (front > 0) ) return SPLIT;
    if (  back  > 0                 ) return BACK;
    if (  front > 0                 ) return FRONT;
    return COPLAN;
 }

/*------------------------------------------------------
 * Fonction de split d'un polygone par rapport a un plan
 *------------------------------------------------------
 */
 static BSP * split( OBJECT_f * obj, POLY_f * p_src, BSP * cur )
 {
    V3D_f    *  ver;                    // tableau des pts
    POLY_f   *  poly;                   // polygone a decouper
    int         count;                  // nb pts du polygone a decouper
    int         out_c = 0;              // nombre de pts en sortie
    int         in_c  = 0;
    V3D_f       ptA;                    // points courants
    V3D_f       ptB;                    // points courant
    float       sideA, sideB;           // calcul
    double      scale;
    int         i;

    V3D_f       k,ext;                  // point intermediaire ( calcul )

    POLY_f   *  new_e;                  // nouveau poly cree
    BSP      *  new_b;

    int         pts[300];               // liste temporaire des points de la face
    int         u[300];                 // liste temporaire des coord u et v de la face
    int         v[300];


    if ( NumCompilePoly < MaxCompilePoly )
    {
       new_e = &obj->poly_begin[NumCompilePoly];
       NumCompilePoly++;
    }
    else BSPERROR("Nombre maximum de polygones atteind",30)

    ver    = obj->vtx_begin;
    poly   = cur->ptpoly;
    count  = poly->nb_vtx;

    // recopie des points
    for ( i = 0 ; i < count ; i++ )
    {
       pts[i] = poly->index[i];
       u[i]   = poly->u[i];
       v[i]   = poly->v[i];
    }

    // copie les infos de la face
    new_e->zmoy = poly->zmoy;
    new_e->coul = poly->coul;
    new_e->nx   = poly->nx;
    new_e->ny   = poly->ny;
    new_e->nz   = poly->nz;
    new_e->state= poly->state;
    new_e->type = poly->type;
    new_e->map  = poly->map;

    ptA.x = ver[pts[count - 1]].x;
    ptA.y = ver[pts[count - 1]].y;
    ptA.z = ver[pts[count - 1]].z;
    ptA.u = u[count - 1];
    ptA.v = v[count - 1];
    sideA = dist_point( obj, p_src, &ptA );

    // test l'intersection de chaque ligne
    for( i = 0 ; i < count ; i++ )
    {
       ptB.x = ver[pts[i]].x;
       ptB.y = ver[pts[i]].y;
       ptB.z = ver[pts[i]].z;
       ptB.u = u[i];
       ptB.v = v[i];
       sideB = dist_point( obj, p_src, &ptB );

       if ( sideB > 0.001 )
       {
          if ( sideA < -0.001 )
          {
             k.x = ptA.x - ptB.x;
             k.y = ptA.y - ptB.y;
             k.z = ptA.z - ptB.z;
             k.u = ptA.u - ptB.u;
             k.v = ptA.v - ptB.v;

             scale   = (double) ( sideA / (double)dot_product_f( k.x, k.y, k.z, p_src->nx, p_src->ny, p_src->nz ));

             ext.x = ptA.x - ( k.x * scale );
             ext.y = ptA.y - ( k.y * scale );
             ext.z = ptA.z - ( k.z * scale );
             ext.u = ptA.u - ( k.u * scale );
             ext.v = ptA.v - ( k.v * scale );

             // ajout du nouveau pt a la liste
             if ( NumCompileVtx < MaxCompileVtx )
             {
                ver[NumCompileVtx].x = ext.x;
                ver[NumCompileVtx].y = ext.y;
                ver[NumCompileVtx].z = ext.z;
                NumCompileVtx++;
             }
             else BSPERROR("Nombre maximum de vertex atteind",31)

             // ajout du point dans la liste de vtx du polygone
             new_e->index[out_c]   = NumCompileVtx - 1;
             new_e->ptvtx3d[out_c] = &ver[NumCompileVtx - 1];
             new_e->u[out_c]       = ext.u;
             new_e->v[out_c]       = ext.v;
             out_c++;

             poly->index[in_c]   = NumCompileVtx - 1;
             poly->ptvtx3d[in_c] = &ver[NumCompileVtx - 1];
             poly->u[in_c]       = ext.u;
             poly->v[in_c]       = ext.v;
             in_c++;
          }
          new_e->index[out_c]   = pts[i];
          new_e->ptvtx3d[out_c] = &ver[pts[i]];
          new_e->u[out_c]       = u[i];
          new_e->v[out_c]       = v[i];
          out_c++;
       }
       else if ( sideB < -0.001 )
            {
               if ( sideA > 0.001 )
               {
                   k.x = ptA.x - ptB.x;
                   k.y = ptA.y - ptB.y;
                   k.z = ptA.z - ptB.z;
                   k.u = ptA.u - ptB.u;
                   k.v = ptA.v - ptB.v;

                   scale   = (double) (sideA / (double) dot_product_f( k.x, k.y, k.z, p_src->nx, p_src->ny, p_src->nz ));

                   ext.x = ptA.x - ( k.x * scale );
                   ext.y = ptA.y - ( k.y * scale );
                   ext.z = ptA.z - ( k.z * scale );
                   ext.u = ptA.u - ( k.u * scale );
                   ext.v = ptA.v - ( k.v * scale );

                  // ajout du nouveau pt a la liste
                  if ( NumCompileVtx < MaxCompileVtx )
                  {
                     ver[NumCompileVtx].x = ext.x;
                     ver[NumCompileVtx].y = ext.y;
                     ver[NumCompileVtx].z = ext.z;
                     NumCompileVtx++;
                  }
                  else BSPERROR("Nombre maximum de vertex atteind",32)

                  // ajout du point dans la liste de vtx du polygone
                  new_e->index[out_c]   = NumCompileVtx - 1;
                  new_e->ptvtx3d[out_c] = &ver[NumCompileVtx - 1];
                  new_e->u[out_c]       = ext.u;
                  new_e->v[out_c]       = ext.v;
                  out_c++;

                  poly->index[in_c]   = NumCompileVtx - 1;
                  poly->ptvtx3d[in_c] = &ver[NumCompileVtx - 1];
                  poly->u[in_c]       = ext.u;
                  poly->v[in_c]       = ext.v;
                  in_c++;
               }
               poly->index[in_c]   = pts[i];
               poly->ptvtx3d[in_c] = &ver[pts[i]];
               poly->u[in_c]       = u[i];
               poly->v[in_c]       = v[i];
               in_c++;
            }
            else
            {
               // le point appartient au plan
               new_e->index[out_c]   = pts[i];
               new_e->ptvtx3d[out_c] = &ver[pts[i]];
               new_e->u[out_c]       = u[i];
               new_e->v[out_c]       = v[i];
               out_c++;

               poly->index[in_c]   = pts[i];
               poly->ptvtx3d[in_c] = &ver[pts[i]];
               poly->u[in_c]       = u[i];
               poly->v[in_c]       = v[i];
               in_c++;
            }
            ptA   = ptB;
            sideA = sideB;
    }
    new_e->nb_vtx = out_c;
    poly->nb_vtx  = in_c;

    // allocation de la nouvelle cellule bsp
    new_b = ( BSP * ) malloc ( sizeof ( BSP ) );
    if ( !new_b ) BSPERROR( "Impossible d'allouer une nouvelle cellule BSP",33);
    new_b->ptpoly = new_e;
    new_b->pnextlst  = NULL;
    new_b->pbackbsp  = NULL;
    new_b->pfrontbsp = NULL;

    return new_b;
}

/*-----------------------------------------------------
 * Fonction pour trouver la racine de l'arbre courant
 * selon le noeud qui coupe le moins d'autre noeuds
 *-----------------------------------------------------
 */
 static float calc_score(float count, float front, float back, float coplan )
 {
    float score;

    score = fabs(front-back) + 10 * count;

    return score;
 }

 static BSP * find_root( OBJECT_f * src, BSP ** tete, int num, int optim )
 {
   float best_score = 9999999.0;
   int   total_test;
   int   i,j;
   float count,coplan;
   float back,front;
   int   res;
   float score   = 0;
   char  work[4] = { '-','/','|','\\' }; // indique qu'on travaille
   int   wcour   = 0;

   BSP * torm    = NULL;           // element a supprimer
   BSP * berm    = NULL;           // element le precedant

   BSP * backlst = NULL;           // second parcour de la liste
   BSP * list    = NULL;           // premier parcour
   BSP * prec    = NULL;           // precedant dans premier parcour
   int   index   = 0;              // index du root dans la liste

   BSP * best_root = NULL;
   POLY_f * cur_root;

   total_test = num;

   if ( optim == NOOPTIM )
   {
      // et c'est parti
      if ( !rootfile )
         if ( total_test > 100 ) total_test = 100;

      for ( list = *tete, i = 0 ; i < total_test ; i++ )
      {
         // affichage comme quoi c'est pas plante
         if ( rootfile )
         {
            printf("%c%c",8,work[ (wcour++)&3 ]);
            fflush(stdout);
         }

         count  = 0.0;
         front  = 0.0;
         back   = 0.0;
         coplan = 0.0;
         // choix du root
         cur_root = list->ptpoly;

         // on compte les coupures
         for ( backlst = *tete, j = 0 ; j < num ; j++, backlst = backlst->pnextlst )
         {
            if ( cur_root != backlst->ptpoly )
            {
               res = classify( src, cur_root, backlst->ptpoly );
               if ( res == SPLIT )  count++;
               if ( res == FRONT )  front++;
               if ( res == BACK )   back++;
               if ( res == COPLAN ) coplan++;
            }
         }

         score = calc_score( count, front, back, coplan );

         // on test
         if ( score < best_score )
         {
            best_score = score;
            best_root  = list;
            torm       = list;
            berm       = prec;
            index      = i;
            if ( best_score == 0.0 )
               break;
         }

         // deplacement dans la liste
         prec = list;
         list = list->pnextlst;
      }
      if ( index != 0 )
         berm->pnextlst = torm->pnextlst;
      else
      {
         berm  = *tete;
         *tete = berm->pnextlst;
      }

      best_root->pfrontbsp = NULL;
      best_root->pbackbsp  = NULL;
      best_root->pnextlst  = NULL;
      // sauvegarde l'index
      if ( rootfile )
      {
         if ( ptrootback < MAXROOT )
         {
            rootback[ptrootback] = index;
            ptrootback++;
         }
         else BSPERROR("Trop de root crees",39)
      }
      if ( rootfile )
      {
        printf("%c %4.1f",8,best_score);
        fflush(stdout);
      }
      return best_root;
   }
   else
   {
      // recuperation du root dans la liste
      total_test = pack_igetw(rootfile);
      for ( list = *tete, i = 0 ; i < total_test ; i++, prec = list, list = list->pnextlst);
      best_root = list;
      torm      = list;
      berm      = prec;

      if ( berm )
         berm->pnextlst = torm->pnextlst;
      else
         *tete = (*tete)->pnextlst;

      best_root->pfrontbsp = NULL;
      best_root->pbackbsp  = NULL;
      best_root->pnextlst  = NULL;
      return best_root;
   }

}


/*-----------------------------------------------------
 * Fonction recursive de construction
 *-----------------------------------------------------
 */
static BSP * Building( OBJECT_f * src, BSP ** list, int num_element, int optim )
{

   int      Sstatus;
   int      Sflag;
   BSP   ** Spfrontpoly;
   BSP   ** Spbackpoly;
   BSP   ** Spnextpoly;
   BSP    * Snew_e;
   BSP    * Scur;

   BSP    * root;                     // la racine courante
   int      nbeltfront;
   int      nbeltback;
   int      fin        = 0;

   BSP    ** temp;                     // un intermediaire
   BSP    ** pile[MAXPILE];            // la pile qui stoque le chainage
   BSP     * pilebsp[MAXPILE];         // la pile qui stoque le root
   int       pileint[MAXPILE];         // la pile pour nbelementback
   int       ptpile     = 0;           // pteur dans le pile

   // c'est parti
   while(!fin)
   {
     if ( optim == NOOPTIM && rootfile )
     {
        printf("\n%d :  ",iter);
        fflush(stdout);
        iter++;
     }

     nbeltfront  = 0;
     nbeltback   = 0;
     // on recherche la partition root
     root        = find_root( src, list, num_element, optim );
     Spfrontpoly = &root->pfrontbsp;
     Spbackpoly  = &root->pbackbsp;
     Spnextpoly  = &root->pnextlst;
     Scur        = *list;
     while ( Scur != NULL )
     {
        *list   = (*list)->pnextlst;
        Sstatus = classify( src, root->ptpoly, Scur->ptpoly );
        Scur->pbackbsp  = NULL;
        Scur->pfrontbsp = NULL;
        Scur->pnextlst  = NULL;
        switch ( Sstatus )
        {
           // construction de la liste en arriere
           case BACK   : *Spbackpoly = Scur;
                         Spbackpoly = &Scur->pnextlst;
                         nbeltback++;
                         break;

           // construction de la liste en avant
           case FRONT  : *Spfrontpoly = Scur;
                         Spfrontpoly = &Scur->pnextlst;
                         nbeltfront++;
                         break;

           // construction de la liste des ploy coplannaires
           case COPLAN : *Spnextpoly = Scur;
                         Spnextpoly = &Scur->pnextlst;
                         break;

           case SPLIT  : //puts("split");
                         Snew_e = split( src, root->ptpoly, Scur );
                         Sflag = classify( src, root->ptpoly, Scur->ptpoly );
                         if ( Sflag == FRONT )
                         {
                            *Spfrontpoly = Scur;
                            Spfrontpoly  = &Scur->pnextlst;
                            nbeltfront++;

                            *Spbackpoly = Snew_e;
                            Spbackpoly  = &Snew_e->pnextlst;
                            nbeltback++;
                         }
                         else
                         {
                            *Spbackpoly = Scur;
                            Spbackpoly  = &Scur->pnextlst;
                            nbeltback++;

                            *Spfrontpoly = Snew_e;
                            Spfrontpoly  = &Snew_e->pnextlst;
                            nbeltfront++;
                         }
                         break;
        }
        Scur   = *list;
     }

     if ( root->pfrontbsp != NULL )
     {
        if (ptpile < MAXPILE )
        {
          pile[ptpile]    = &root->pfrontbsp;
          pilebsp[ptpile] = root;
          pileint[ptpile] = nbeltback;
          ptpile++;
        }
        else BSPERROR("Pile de construction BSP pleine",36)

        *list           = root->pfrontbsp;
        num_element     = nbeltfront;
     }
     else
     {
        if ( nbeltback != 0 )
        {
           if (ptpile < MAXPILE )
           {
              pile[ptpile]    = &root->pbackbsp;
              pilebsp[ptpile] = root;
              pileint[ptpile] = 0;
              ptpile++;
           }
           else BSPERROR("Pile de construction BSP pleine",36)
           *list           = root->pbackbsp;
           num_element     = nbeltback;
        }
        else
        {
           while ( nbeltback == 0 && ptpile != 0)
           {
              ptpile--;
              nbeltback = pileint[ptpile];
              temp  = pile[ptpile];
              *temp = root;
              root  = pilebsp[ptpile];
           }
           if ( ptpile != 0 )
           {
              if( root->pbackbsp != NULL )
              {
                 if (ptpile < MAXPILE )
                 {
                    pile[ptpile]    = &root->pbackbsp;
                    pilebsp[ptpile] = root;
                    pileint[ptpile] = 0;
                    ptpile++;
                 }
                 else BSPERROR("Pile de construction BSP pleine",36)
                 *list           = root->pbackbsp;
                 num_element     = nbeltback;
              }
              else fin = 1;
           }
           else fin = 1;
        }
     }
   }
   return root;

}
/*-----------------------------------------------------
 * supprime les points doublons puisqu'il y en a
 *-----------------------------------------------------
 */
 static void del_points( OBJECT_f * ob )
 {
    int       i,j;
    POLY_f  * poly;
    V3D_f   * pa;
    V3D_f   * pb;
    V3D_f     v1;               // optimisation le vect unit n'est calcule qu'une fois
    int       count;

    int       outc = 0;         // compteur

    for ( i = 0 ; i < ob->nb_poly_begin ; i++ )
    {
       outc  = 0;
       poly  = &ob->poly_begin[i];
       count = poly->nb_vtx;
       pa    = poly->ptvtx3d[count-1];

          // test chaque ligne
       for( j = 0 ; j < count ; j++ )
       {
          pb   = poly->ptvtx3d[j];

          v1.x = pa->x - pb->x;
          v1.y = pa->y - pb->y;
          v1.z = pa->z - pb->z;
          if ( v1.x != 0 || v1.y != 0 || v1.z != 0 )
          {
             poly->index[outc]   = poly->index[j];
             poly->ptvtx3d[outc] = poly->ptvtx3d[j];
             poly->u[outc]       = poly->u[j];
             poly->v[outc]       = poly->v[j];
             outc++;
          }
          pa = pb;
       }
       poly->nb_vtx = outc;
    }
}



/*-----------------------------------------------------
 * Ajout des nouveaux points aux polygones qui passe
 * par ces points pour eviter les traits noir
 *-----------------------------------------------------
 */
 static float give_dist( V3D_f * p1, V3D_f * p2, V3D_f * vn, V3D_f * p3 )
 {

    V3D_f v1,v2;
    float dist;

    // un point p3 est compris dans le segment [p1 p2] si
    // il est entre les deux points ou si c'est lui
    // et si sont vecteur avec un point est parallele
    // a celui entre les deux points

    if ( (p3->x == p1->x && p3->y == p1->y && p3->z == p1->z) ||
         (p3->x == p2->x && p3->y == p2->y && p3->z == p2->z)
       )
       // le point est un de ceux la, on l'elimine
       return -1.0;

    if (    ( (p3->x >= p1->x && p3->x <= p2->x)  || (p3->x >= p2->x && p3->x <= p1->x) )
         && ( (p3->y >= p1->y && p3->y <= p2->y)  || (p3->y >= p2->y && p3->y <= p1->y) )
         && ( (p3->z >= p1->z && p3->z <= p2->z)  || (p3->z >= p2->z && p3->z <= p1->z) )
       )
    {
       // le point est entre les deux autres, on verifie qu'il soit sur la meme ligne
       v2.x = p1->x - p3->x;
       v2.y = p1->y - p3->y;
       v2.z = p1->z - p3->z;
       normalize_vector_f( &v2.x, &v2.y, &v2.z );
       // si les vecteur sont colineaire : produit vectoriel -> 1.0
       dist = dot_product_f( vn->x, vn->y, vn->z, v2.x, v2.y, v2.z );
       if ( dist >= 0.999 && dist <= 1.001 )
       {
          // ce point appartient au segment calculer sa distance
          v2.x = p1->x - p3->x;
          v2.y = p1->y - p3->y;
          v2.z = p1->z - p3->z;

          v1.x = p1->x - p2->x;
          v1.y = p1->y - p2->y;
          v1.z = p1->z - p2->z;
          if ( v1.x == 0.0 && v1.y == 0.0 && v1.z == 0.0 )
             BSPERROR("Vecteur bizard",102);
          dist = sqrt( v2.x * v2.x + v2.y * v2.y + v2.z * v2.z ) /
                 sqrt( v1.x * v1.x + v1.y * v1.y + v1.z * v1.z );
          return dist;
       }
       else
          return -1.0;
    }
    else
       return -1.0;
 }


 static void add_points( OBJECT_f * ob, int optim )
 {
    int       i,j,k,l;
    POLY_f  * poly;
    V3D_f   * pa;
    V3D_f   * pb;
    V3D_f   * pc;
    V3D_f     v1;               // optimisation le vect unit n'est calcule qu'une fois
    int       count;
    int       min;
    int       tmpi;
    float     tmpf,tmpu,tmpv;
    float     dist;
    float     ua,va,ub,vb;

    float     tabd[300];        // stokage du nouvel ordre par les distances
    int       tabi[300];        // idem index
    float     tabu[300];
    float     tabv[300];
    int       inter = 0;        // points intermediaires

    int       tab[300];         // resultats finaux
    float     tabu1[300];
    float     tabv1[300];
    int       outc = 0;         // compteur



    if ( optim == NOOPTIM )
    {
       printf("\n Ajout des nouveaux points aux faces :       ");
       fflush(stdout);
       for ( i = 0 ; i < ob->nb_poly_begin ; i++ )
       {
          printf("%c%c%c%c%c%c%4.1f %%",8,8,8,8,8,8,(float)100.0*i/ob->nb_poly_begin);
          fflush(stdout);
          outc  = 0;
          poly  = &ob->poly_begin[i];
          count = poly->nb_vtx;
          pa    = poly->ptvtx3d[count-1];
          ua    = poly->u[count-1];
          va    = poly->v[count-1];

          // test chaque ligne
          for( j = 0 ; j < count ; j++ )
          {
             pb    = poly->ptvtx3d[j];
             ub    = poly->u[j];
             vb    = poly->v[j];
             inter = 0;

             v1.x = pa->x - pb->x;
             v1.y = pa->y - pb->y;
             v1.z = pa->z - pb->z;
             normalize_vector_f( &v1.x, &v1.y, &v1.z );

             // test chaque point ajoute
             for ( k = FirstAddedVtx ; k < NumCompileVtx ; k++ )
             {
                pc = &ob->vtx_begin[k];
                dist = give_dist( pa, pb, &v1, pc );
                if ( dist > 0.0 )
                {
                   // on ajoute ce point a la liste
                   tabd[inter] = dist;
                   tabi[inter] = k;
                   tabu[inter] = ( ua - ub ) * dist;
                   tabv[inter] = ( va - vb ) * dist;
                   inter++;
                }
             }

             // classe les points
             if ( inter > 0 )
             {
                for ( k = 0 ; k < inter-1 ; k++ )
                {
                   min = k;
                   for ( l = k+1 ; l < inter ; l++ )
                      if ( tabd[l] < tabd[min] ) min = l;
                   tmpu = tabu[k];
                   tmpv = tabv[k];
                   tmpi = tabi[k];
                   tmpf = tabd[k];
                   tabi[k] = tabi[min];
                   tabd[k] = tabd[min];
                   tabu[k] = tabu[min];
                   tabv[k] = tabv[min];
                   tabi[min] = tmpi;
                   tabd[min] = tmpf;
                   tabu[min] = tmpu;
                   tabv[min] = tmpv;
                }

                // ajoute les points
                for ( k = 0 ; k < inter ; k++ )
                {
                   tab[outc]   = tabi[k];
                   tabu1[outc] = tabu[k];
                   tabv1[outc] = tabv[k];
                   outc++;
                }
             }

             // ajoute pb
             tab[outc]   = poly->index[j];
             tabu1[outc] = poly->u[j];
             tabv1[outc] = poly->v[j];
             outc++;

             // la suite
             pa = pb;
             ua = ub;
             va = vb;
          }

          // recopie la nouvelle liste de point dans le poly
          if ( outc > poly->nb_vtx )
          {
             vtxmerged[ptvtxmerged] = i;
             ptvtxmerged++;
             for ( j = 0 ; j < outc ; j++ )
             {
                poly->index[j]   = tab[j];
                poly->u[j]       = tabu1[j];
                poly->v[j]       = tabv1[j];
                poly->ptvtx3d[j] = &ob->vtx_begin[ tab[j] ];
             }
             poly->nb_vtx = outc;
          }
       }
       printf("\n");
    }
    else
    {
       int nbpoly;
       int nopoly;

       nbpoly = pack_igetw(rootfile);
       for ( i = 0 ; i < nbpoly ; i++ )
       {
          nopoly = pack_igetw(rootfile);
          outc   = 0;
          poly   = &ob->poly_begin[nopoly];
          count  = poly->nb_vtx;
          pa     = poly->ptvtx3d[count-1];
          ua     = poly->u[count-1];
          va     = poly->v[count-1];

          // test chaque ligne
          for( j = 0 ; j < count ; j++ )
          {
             pb    = poly->ptvtx3d[j];
             ub    = poly->u[j];
             vb    = poly->v[j];
             inter = 0;

             v1.x = pa->x - pb->x;
             v1.y = pa->y - pb->y;
             v1.z = pa->z - pb->z;
             normalize_vector_f( &v1.x, &v1.y, &v1.z );

             // test chaque point ajoute
             for ( k = FirstAddedVtx ; k < NumCompileVtx ; k++ )
             {
                pc = &ob->vtx_begin[k];
                dist = give_dist( pa, pb, &v1, pc );
                if ( dist > 0.0 )
                {
                   // on ajoute ce point a la liste
                   tabd[inter] = dist;
                   tabi[inter] = k;
                   tabu[inter] = ( ua - ub ) * dist;
                   tabv[inter] = ( va - vb ) * dist;
                   inter++;
                }
             }

             // classe les points
             if ( inter > 0 )
             {
                for ( k = 0 ; k < inter-1 ; k++ )
                {
                   min = k;
                   for ( l = k+1 ; l < inter ; l++ )
                      if ( tabd[l] < tabd[min] ) min = l;
                   tmpu = tabu[k];
                   tmpv = tabv[k];
                   tmpi = tabi[k];
                   tmpf = tabd[k];
                   tabi[k] = tabi[min];
                   tabd[k] = tabd[min];
                   tabu[k] = tabu[min];
                   tabv[k] = tabv[min];
                   tabi[min] = tmpi;
                   tabd[min] = tmpf;
                   tabu[min] = tmpu;
                   tabv[min] = tmpv;
                }

                // ajoute les points
                for ( k = 0 ; k < inter ; k++ )
                {
                   tab[outc]   = tabi[k];
                   tabu1[outc] = tabu[k];
                   tabv1[outc] = tabv[k];
                   outc++;
                }
             }

             // ajoute pb
             tab[outc]   = poly->index[j];
             tabu1[outc] = poly->u[j];
             tabv1[outc] = poly->v[j];
             outc++;

             // la suite
             pa = pb;
             ua = ub;
             va = vb;
          }

          // recopie la nouvelle liste de point dans le poly
          if ( outc > poly->nb_vtx )
          {
             for ( j = 0 ; j < outc ; j++ )
             {
                poly->index[j]   = tab[j];
                poly->u[j]       = tabu1[j];
                poly->v[j]       = tabv1[j];
                poly->ptvtx3d[j] = &ob->vtx_begin[ tab[j] ];
             }
             poly->nb_vtx = outc;
          }
       }
    }

}


/*-----------------------------------------------------
 * Construction de l'arbre BSP
 * Construction d'un fichier de root
 *  NULL pour pas en creer
 *-----------------------------------------------------
 */
  BSP * Build_Bsp3d( OBJECT_f * src, OBJECT_f ** dest, int max_vtx, int max_poly, int vtxbypoly, char * file )
  {
     int         i,j;
     BSP      *  liste;
     BSP      *  arbre;
     BSP      *  prec;
     OBJECT_f *  dst;

     lightify_object_f( src );

     // creation d'un fichier pour sauver les root
     if ( file )
     {
        rootfile = pack_fopen(file,"wp");
        if ( !rootfile ) BSPERROR("Impossible de creer le fichier de root",38)
        rootback = ( int * ) malloc( MAXROOT * sizeof(int) );
        if ( !rootback ) BSPERROR("Impossible d'allouer la memoire a la file root",100);
        ptvtxmerged = 0;
     }
     else
        rootfile = NULL;

     // init
     MaxCompilePoly = max_poly;
     MaxCompileVtx  = max_vtx;
     NumCompilePoly = src->nb_poly_begin;
     NumCompileVtx  = src->nb_vtx;
     FirstAddedVtx  = src->nb_vtx;

     // creation d'un nouvel objet de taille adequat
     dst = create_object_f( max_vtx, max_poly, vtxbypoly+10 );

     dst->nb_poly_begin = src->nb_poly_begin;
     dst->nb_vtx        = src->nb_vtx;

     // recopie des infos de l'objet source
     for ( i = 0 ; i < src->nb_vtx ; i++ )
         dst->vtx_begin[i] = src->vtx_begin[i];
     for ( i = 0 ; i < src->nb_poly_begin ; i++ )
     {
         dst->poly_begin[i].nb_vtx = src->poly_begin[i].nb_vtx;
         dst->poly_begin[i].zmoy   = src->poly_begin[i].zmoy;
         dst->poly_begin[i].coul   = src->poly_begin[i].coul;
         dst->poly_begin[i].nx     = src->poly_begin[i].nx;
         dst->poly_begin[i].ny     = src->poly_begin[i].ny;
         dst->poly_begin[i].nz     = src->poly_begin[i].nz;
         dst->poly_begin[i].state  = src->poly_begin[i].state;
         dst->poly_begin[i].type   = src->poly_begin[i].type;
         dst->poly_begin[i].map    = src->poly_begin[i].map;
         for ( j = 0 ; j < vtxbypoly ; j++ )
         {
            dst->poly_begin[i].u[j]       = src->poly_begin[i].u[j];
            dst->poly_begin[i].v[j]       = src->poly_begin[i].v[j];
            dst->poly_begin[i].index[j]   = src->poly_begin[i].index[j];
            dst->poly_begin[i].ptvtx3d[j] = src->poly_begin[i].ptvtx3d[j];
         }
     }

     // destruction de l'objet source
     destroy_object_f(src);
     src = NULL;

     // creation de la liste BSP
     prec = NULL;
     liste = ( BSP * ) malloc ( dst->nb_poly_begin * sizeof ( BSP ) );
     if ( !liste )
        BSPERROR("Impossible d'allouer la liste BSP",35)

     for ( i = 0 ; i < dst->nb_poly_begin ; i++ )
     {
        liste[i].ptpoly    = &dst->poly_begin[i];
        liste[i].pnextlst  = prec;
        liste[i].pfrontbsp = NULL;
        liste[i].pbackbsp  = NULL;
        prec               = &liste[i];
     }
     liste = prec;

     // construction de l'arbre BSP
     arbre = Building( dst, &liste, dst->nb_poly_begin, NOOPTIM );

     // mise a jour des nombre de poly et vtx
     dst->nb_vtx = NumCompileVtx;
     dst->nb_poly_begin = NumCompilePoly;

     // supprime les points doublon dans chaque face
     del_points( dst );
     // merge les nouveaux points
     add_points( dst, NOOPTIM );

     // ecriture des infos
     if ( rootfile )
     {
        pack_iputw(dst->nb_vtx,rootfile);
        pack_iputw(dst->nb_poly_begin,rootfile);
        pack_iputw(vtxbypoly,rootfile);
        for( i = 0 ; i < ptrootback ; i++ )
           pack_iputw(rootback[i],rootfile);

        pack_iputw(ptvtxmerged,rootfile);
        for( i = 0 ; i < ptvtxmerged ; i++ )
           pack_iputw(vtxmerged[i],rootfile);
        pack_fclose(rootfile);
        free(rootback);
     }

     // retourne l'arbre
     *dest = dst;
     return arbre;
}

/*-----------------------------------------------------
 * Construction de l'arbre BSP avec fichier
 * de choix des root
 *-----------------------------------------------------
 */
  BSP * Build_Bsp3d_optim( OBJECT_f * src, OBJECT_f ** dest, char * file )
  {
     int         i,j;
     BSP      *  liste;
     BSP      *  arbre;
     BSP      *  prec;
     OBJECT_f *  dst;
     int         max_vtx;
     int         max_poly;
     int         vtxbypoly;

     // optimise la taille de l'objet source
     lightify_object_f( src );

     // ouverture du fichier des root
     rootfile  = NULL;
     rootfile  = pack_fopen(file,"rp");
     if (!rootfile) BSPERROR("Impossible d'ouvrir le fichier d'optimisation",37)

     // recuperation infos de depart
     max_vtx   = pack_igetw(rootfile);
     max_poly  = pack_igetw(rootfile);
     vtxbypoly = pack_igetw(rootfile);

     MaxCompilePoly = max_poly;
     MaxCompileVtx  = max_vtx;
     NumCompilePoly = src->nb_poly_begin;
     NumCompileVtx  = src->nb_vtx;
     FirstAddedVtx  = src->nb_vtx;

     // creation d'un nouvel objet de taille adequat
     dst = create_object_f( max_vtx, max_poly, vtxbypoly+10 );

     dst->nb_poly_begin = src->nb_poly_begin;
     dst->nb_vtx        = src->nb_vtx;

     // recopie des infos de l'objet source
     for ( i = 0 ; i < src->nb_vtx ; i++ )
         dst->vtx_begin[i] = src->vtx_begin[i];
     for ( i = 0 ; i < src->nb_poly_begin ; i++ )
     {
         dst->poly_begin[i].nb_vtx = src->poly_begin[i].nb_vtx;
         dst->poly_begin[i].zmoy   = src->poly_begin[i].zmoy;
         dst->poly_begin[i].coul   = src->poly_begin[i].coul;
         dst->poly_begin[i].nx     = src->poly_begin[i].nx;
         dst->poly_begin[i].ny     = src->poly_begin[i].ny;
         dst->poly_begin[i].nz     = src->poly_begin[i].nz;
         dst->poly_begin[i].state  = src->poly_begin[i].state;
         dst->poly_begin[i].type   = src->poly_begin[i].type;
         dst->poly_begin[i].map    = src->poly_begin[i].map;
         for ( j = 0 ; j < vtxbypoly ; j++ )
         {
            dst->poly_begin[i].u[j]       = src->poly_begin[i].u[j];
            dst->poly_begin[i].v[j]       = src->poly_begin[i].v[j];
            dst->poly_begin[i].index[j]   = src->poly_begin[i].index[j];
            dst->poly_begin[i].ptvtx3d[j] = src->poly_begin[i].ptvtx3d[j];
         }
     }

     // destruction de l'objet source
     destroy_object_f(src);
     src = NULL;

     // creation de la liste BSP
     prec = NULL;
     liste = ( BSP * ) malloc ( dst->nb_poly_begin * sizeof ( BSP ) );
     if ( !liste )
        BSPERROR("Impossible d'allouer la liste BSP",35)

     for ( i = 0 ; i < dst->nb_poly_begin ; i++ )
     {
        liste[i].ptpoly    = &dst->poly_begin[i];
        liste[i].pnextlst  = prec;
        liste[i].pfrontbsp = NULL;
        liste[i].pbackbsp  = NULL;
        prec               = &liste[i];
     }
     liste = prec;

     // construction de l'arbre BSP
     arbre = Building( dst, &liste, dst->nb_poly_begin, OPTIM );

     // mise a jour des nombre de poly et vtx
     dst->nb_vtx = NumCompileVtx;
     dst->nb_poly_begin = NumCompilePoly;

     // supprime les points doublon dans chaque face
     del_points( dst );
     // suppression des zones noires
     add_points( dst, OPTIM );

     pack_fclose(rootfile);
     // retourne l'arbre
     *dest = dst;
     return arbre;
}
/*=============================================================================
 * PARCOUR DE L'ARBRE BSP
 *=============================================================================
 */

 // plan de coupure arriere
 static float p_back_nx,p_back_ny,p_back_nz;
 static float p_obs_x,p_obs_y,p_obs_z;
 // plan de coupure droit
 static float p_right_nx,p_right_ny,p_right_nz;
 // plan de coupure gauche
 static float p_left_nx,p_left_ny,p_left_nz;
 // plan de coupure up
 static float p_up_nx,p_up_ny,p_up_nz;
 // plan de codownure down
 static float p_down_nx,p_down_ny,p_down_nz;

 static V3D_f obs;
 static float p_zmax;

 // les points limitant le domaine de visualisation
 static float p_pt1_x,p_pt1_y,p_pt1_z;
 static float p_pt2_x,p_pt2_y,p_pt2_z;
 static float p_pt3_x,p_pt3_y,p_pt3_z;
 static float p_pt4_x,p_pt4_y,p_pt4_z;


/*----------------------------------------------
 * mise a jour des plan de coupure
 *----------------------------------------------
 */

 void Update_Plan( float zmax, float tx, float ty, float tz, float xrot, float yrot, float zrot, float scale )
 {
     MATRIX_f m1,m2,m3,m4,m5;
     // plan arriere
     float    back_nx = 0.0;
     float    back_ny = 0.0;
     float    back_nz = 1.0;
     float    obs_x  = 0.0;
     float    obs_y  = 0.0;
     float    obs_z  = -1.5;
     // plan droit
     float    right_nx = -0.707;
     float    right_ny =  0.0;
     float    right_nz =  0.707;
     // plan gauche
     float    left_nx =  0.707;
     float    left_ny =  0.0;
     float    left_nz =  0.707;
     // plan haut
     float    up_nx =  0.0;
     float    up_ny =  0.707;
     float    up_nz =  0.707;
     // plan bas
     float    down_nx = 0.0;
     float    down_ny = -0.707;
     float    down_nz = 0.707;
     // les points limitant le domaine visible
     float    p1_x = -zmax;
     float    p1_y =  zmax;
     float    p1_z =  zmax;

     float    p2_x =  zmax;
     float    p2_y = -zmax;
     float    p2_z =  zmax;

     float    p3_x =  zmax;
     float    p3_y =  zmax;
     float    p3_z =  zmax;

     float    p4_x = -zmax;
     float    p4_y = -zmax;
     float    p4_z =  zmax;

     p_zmax = zmax;

     get_transformation_matrix_f(&m1, 1.0, 255.0-xrot, 0.0,  0.0, 0.0, 0.0, 0.0);
     get_transformation_matrix_f(&m2, 1.0, 0.0, 255.0-yrot,  0.0, 0.0, 0.0, 0.0);
     get_transformation_matrix_f(&m3, 1.0, 0.0, 0.0, 255.0-zrot,  0.0, 0.0, 0.0);
     matrix_mul_f( &m1, &m2, &m4 );
     matrix_mul_f( &m4, &m3, &m5 );

     // modifie les coordonnees du point du plan
     obs_x = -tx;
     obs_y = -ty;
     if ( scale != 0.0 ) obs_z -= (tz/scale);
     apply_matrix_f( &m5, obs_x,obs_y,obs_z, &p_obs_x,&p_obs_y,&p_obs_z );
     obs.x = p_obs_x;
     obs.y = p_obs_y;
     obs.z = p_obs_z;

     // modifie les points de delimitation de l'espace
     p1_x -= tx;
     p1_y -= ty;
     if ( scale != 0.0 ) p1_z -= (tz/scale);
     p2_x -= tx;
     p2_y -= ty;
     if ( scale != 0.0 ) p2_z -= (tz/scale);
     p3_x -= tx;
     p3_y -= ty;
     if ( scale != 0.0 ) p3_z -= (tz/scale);
     p4_x -= tx;
     p4_y -= ty;
     if ( scale != 0.0 ) p4_z -= (tz/scale);


     // calcul des nouvelles coord du plan dans l'espace objet
     apply_matrix_f( &m5, back_nx,back_ny,back_nz, &p_back_nx,&p_back_ny,&p_back_nz );
     apply_matrix_f( &m5, right_nx,right_ny,right_nz, &p_right_nx,&p_right_ny,&p_right_nz );
     apply_matrix_f( &m5, left_nx,left_ny,left_nz, &p_left_nx,&p_left_ny,&p_left_nz );
     apply_matrix_f( &m5, up_nx,up_ny,up_nz, &p_up_nx,&p_up_ny,&p_up_nz );
     apply_matrix_f( &m5, down_nx,down_ny,down_nz, &p_down_nx,&p_down_ny,&p_down_nz );

     // calcul les nouvelles coord des points limites du domaine visible
     apply_matrix_f( &m5, p1_x,p1_y,p1_z, &p_pt1_x,&p_pt1_y,&p_pt1_z );
     apply_matrix_f( &m5, p2_x,p2_y,p2_z, &p_pt2_x,&p_pt2_y,&p_pt2_z );
     apply_matrix_f( &m5, p3_x,p3_y,p3_z, &p_pt3_x,&p_pt3_y,&p_pt3_z );
     apply_matrix_f( &m5, p4_x,p4_y,p4_z, &p_pt4_x,&p_pt4_y,&p_pt4_z );

 }

/*------------------------------------------------------
 * Indique si le plan d'une face croise le domaine
 * visible a partir de 5 points limitant ce domaine
 *------------------------------------------------------
 */
 static int is_in_worldwiew( POLY_f * x )
 {
    int   front = 0;
    int   back  = 0;
    float cur;
    float X,Y,Z;
    float px,py,pz;

    px = x->ptvtx3d[0]->x;
    py = x->ptvtx3d[0]->y;
    pz = x->ptvtx3d[0]->z;

    // test la position des points limite du domaine visible par rapport au plan courant

    X = p_pt1_x - px;
    Y = p_pt1_y - py;
    Z = p_pt1_z - pz;
    cur = dot_product_f(X, Y, Z, x->nx, x->ny, x->nz);
    if (  cur > -0.001 ) front++;
    else if ( cur < -0.001 ) back++;

    X = p_pt2_x - px;
    Y = p_pt2_y - py;
    Z = p_pt2_z - pz;
    cur = dot_product_f(X, Y, Z, x->nx, x->ny, x->nz);
    if (  cur > -0.001 ) front++;
    else if ( cur < -0.001 ) back++;

    X = p_pt3_x - px;
    Y = p_pt3_y - py;
    Z = p_pt3_z - pz;
    cur = dot_product_f(X, Y, Z, x->nx, x->ny, x->nz);
    if (  cur > -0.001 ) front++;
    else if ( cur < -0.001 ) back++;

    X = p_pt4_x - px;
    Y = p_pt4_y - py;
    Z = p_pt4_z - pz;
    cur = dot_product_f(X, Y, Z, x->nx, x->ny, x->nz);
    if (  cur > -0.001 ) front++;
    else if ( cur < -0.001 ) back++;

    X = p_obs_x - px;
    Y = p_obs_y - py;
    Z = p_obs_z - pz;
    cur = dot_product_f(X, Y, Z, x->nx, x->ny, x->nz);
    if (  cur > -0.001 ) front++;
    else if ( cur < -0.001 ) back++;

    // on fait la synthese des resultats : tous les points du meme cote = hors domaine
    if ( back  == 5 ) return BACK;
    if ( front == 5 ) return BACK;
    return FRONT;

 }



/*------------------------------------------------------
 * Indique si une face est dans le domaine visible
 * devant l'observateur
 * renvoie FRONT si c'est le cas
 *------------------------------------------------------
 */
 static int is_visible( POLY_f * x )
 {
    int   back0 = 0;
    int   back1 = 0;
    int   back2 = 0;
    int   back3 = 0;
    int   back4 = 0;
    int   back5 = 0;

    int   i;
    float cur;
    float TX[200],TY[200],TZ[200];

    // elimine les faces cachee sens inverse de obs
    TX[0] = x->ptvtx3d[0]->x - p_obs_x;
    TY[0] = x->ptvtx3d[0]->y - p_obs_y;
    TZ[0] = x->ptvtx3d[0]->z - p_obs_z;
    cur = dot_product_f(x->nx, x->ny, x->nz, TX[0], TY[0], TZ[0]);
    if ( cur > 0.002 ) return BACK;


    // elimine a part les faces de derriere et coupure sur zmax
    for ( i = 0 ; i < x->nb_vtx ; i++ )
    {
       TX[i] = x->ptvtx3d[i]->x - p_obs_x;
       TY[i] = x->ptvtx3d[i]->y - p_obs_y;
       TZ[i] = x->ptvtx3d[i]->z - p_obs_z;
       // plan arriere
       cur = dot_product_f(TX[i], TY[i], TZ[i], p_back_nx, p_back_ny, p_back_nz);
       if ( cur <= 0.0 ) back1++;
       else if ( cur > p_zmax ) back0++;
    }
    if ( back1 == x->nb_vtx ) return BACK;      // derriere obs
    if ( back0 == x->nb_vtx ) return BACK;      // trop loin


    // elimine gauche et droit
    for ( i = 0 ; i < x->nb_vtx ; i++ )
    {
       cur = dot_product_f(TX[i], TY[i], TZ[i], p_right_nx, p_right_ny, p_right_nz);
       if ( cur <= 0.0 ) back2++;
       else
       {
          cur = dot_product_f(TX[i], TY[i], TZ[i], p_left_nx, p_left_ny, p_left_nz);
          if ( cur <= 0.0 ) back3++;
       }
    }
    if ( back2 == x->nb_vtx ) return BACK;      // trop a droite
    if ( back3 == x->nb_vtx ) return BACK;      // trop a gauche


    // elimine les plan haut et bas
    for ( i = 0 ; i < x->nb_vtx ; i++ )
    {
       cur = dot_product_f(TX[i], TY[i], TZ[i], p_up_nx, p_up_ny, p_up_nz);
       if ( cur <= 0.0 ) back4++;
       else
       {
          cur = dot_product_f(TX[i], TY[i], TZ[i], p_down_nx, p_down_ny, p_down_nz);
          if ( cur <= 0.0 ) back5++;
       }
    }
    if ( back4 == x->nb_vtx ) return BACK;      // trop haute
    if ( back5 == x->nb_vtx ) return BACK;      // trop basse

    if ( back1 == 0 && back2 == 0 && back3 == 0 && back4 == 0 && back5 == 0 )
       return FRONT;
    else
       return SPLIT;
 }


/*-------------------------------------------------------
 * Construction de l'objet affichable a partir de l'arbre
 * BSP et du points ou on est dans l'espace
 *-------------------------------------------------------
 */
  int Draw_Bsp3d( BSP * arbre, OBJECT_f * src, int maxvtx )
  {
     BSP      * pilebsp[MAXPILE];         // la pile qui stoque le root
     float      piledist[MAXPILE];        // la pile des distances
     int        pilevis[MAXPILE];         // la pile indiquant si la face est visible
     int        ptpile     = 0;           // pteur dans le pile
     float      dist;
     int        fin        = 0;
     int        courpoly   = 0;           // poly a inserer
     OBJECT_f * dst;
     POLY_f   * poly, *poly2;
     int        j;
     int        etat;

     BSP     * cour, *temp;

     dst  = src;
     cour = arbre;

     while ( !fin )
     {
           // on descend tout les back
           while ( cour )
           {
              dist = dist_point( src, cour->ptpoly, &obs );
              if ( ptpile < MAXPILE )
              {
                 pilebsp[ptpile]  = cour;
                 piledist[ptpile] = dist;
                 pilevis[ptpile]  = 1;               // visible par defaut
                 ptpile++;
              }
              else BSPERROR("pile trace pleine",101)

              if ( is_in_worldwiew( cour->ptpoly ) == FRONT )
              {
                 // la face est dans le monde, on la descend
                 if ( dist > 0.0 )
                     cour = cour->pbackbsp;
                 else
                     cour = cour->pfrontbsp;
              }
              else
              {
                 // le plan de la face est hors monde stop a la descente
                 pilevis[ptpile-1] = 0;
                 cour = NULL;
              }
           }

           while ( !cour && !fin )
           {
              if ( ptpile > 0 )
              {
                 // recopie la face et celle qui st coplannaires si le plan est in world
                 ptpile--;
                 cour  = pilebsp[ptpile];
                 dist  = piledist[ptpile];
                 temp  = cour;
                 while ( temp && pilevis[ptpile] == 1 )
                 {
                    poly = temp->ptpoly;
                    etat = is_visible( temp->ptpoly );
                    if ( etat >= FRONT || !poly->state )
                    {
                       // recopie infos dans polyend et poly2d
                       poly2 = &dst->poly_end[courpoly];
                       poly2->nb_vtx = poly->nb_vtx;
                       poly2->coul   = poly->coul;
                       if ( etat == SPLIT && poly->state == 1)
                          poly2->state = 1;
                       else
                          poly2->state = 0;
                       for ( j = 0 ; j < poly->nb_vtx ; j++ )
                       {
                          poly2->index[j]   = poly->index[j];
                          poly2->ptvtx3d[j] = &dst->vtx_end[ poly->index[j] ];
                          poly2->ptvtx2d[j] = &dst->vtx_2d [ poly->index[j] ];
                       }
                       courpoly++;
                    }
                    temp = temp->pnextlst;
                 }

                 // de l'autre cote
                 if ( dist > 0.0 )
                    cour = cour->pfrontbsp;
                 else
                    cour = cour->pbackbsp;

              }
              else fin = 1;
           }
     }
     dst->nb_poly_end  = courpoly;
     return courpoly;
  }


#undef MAXPILE
#undef OPTIM
#undef NOOPTIM
#undef MAXROOT


