/*==========================================================
 * Clipping 3d pour moteur Disk91
 *  Base : algo Abrash Zen de la programmation graphique
 *         p 833  et clip3d allegro
 *----------------------------------------------------------
 * (c) Disk91 / TeKnO-BeDoUiNs / 12 - 1998
 *==========================================================
 */
 #include <allegro.h>
 #include "3d.h"


 #define INT_NONE        0
 #define INT_1COL        1
 #define INT_3COL        2
 #define INT_UV          4




 /*--------------------------------------------------------
  * in est la liste des points en 3d avec de la place
  *   en plus
  * out est la liste des points en 2d avec de la place
  *   pour de nouveaux points ajoutes
  * renvoie le nombre de point de la nouvelle face si > 3
  * renvoie le nombre de vtx ajoutes dans nbad
  *--------------------------------------------------------
  */
  int clip_f( int type, float min_z, int * nbad, int vc, V3D_f ** in, V3D_f ** out )
  {
     V3D_f  ** tmpin;
     V3D_f  ** tmpout;

     int       i;
     int       add;             // index pour ajout de point
     int       courant;         // index pour les points existants
     int       next;            // index pour sommet suivant
     int       sortie;          // index courant de sortie

     int       test1,test2;     // backup des tests
     int       base;            // nombre de vtx a la base
     int       r,b,g;           // interpolation couleurs
     int       color;           // couleur vtx 0 pour flat

     float     scale;
     int       flags;           // pour le type des texture

     static int flag_table[] =
     {
        INT_NONE,                             /* flat */
        INT_1COL,                             /* gcol */
        INT_3COL,                             /* grgb */
        INT_UV,                               /* atex */
        INT_UV,                               /* ptex */
        INT_UV,                               /* atex mask */
        INT_UV,                               /* ptex mask */
        INT_UV + INT_1COL,                    /* atex lit */
        INT_UV + INT_1COL,                    /* ptex lit */
        INT_UV + INT_1COL,                    /* atex mask lit */
        INT_UV + INT_1COL                     /* ptex mask lit */
     };

     flags = flag_table[type];
     color = in[0]->c;

     // allocation d'un buffer temporaire des faces
     tmpin  = (V3D_f **) malloc ( vc * sizeof ( V3D_f * ) );
     tmpout = (V3D_f **) malloc ( vc * sizeof ( V3D_f * ) );
     base   = vc;

/*-----------------------------------------------------------------------------
 * test sur z < min_z
 *-----------------------------------------------------------------------------
 */
     for ( i = 0 ; i < vc ; i++ ) tmpout[i] = out[i];
     for ( i = 0 ; i < vc ; i++ ) tmpin[i]  = in[i];

     // test z < min_z
     courant = 0;
     add     = vc;
     sortie  = 0;

     test1   = (tmpin[courant]->z > min_z);
     for ( i = 0 ; i < vc ; i++ )
     {
        next = (i == vc-1) ? 0 : i+1;

        // on garde le sommet courant s'il est visible
        if ( test1 )
        {
           out[sortie] = tmpout[courant];
           in[sortie]  = tmpin[courant];
           sortie++;
        }

        test2 = ( tmpin[next]->z > min_z );

        // ajouter un sommet couper si une bordure est dans l'ecran et l'autre hors
        if ( test1 != test2 )
        {
           // calcul des nouvelles valeurs
           scale = ( min_z - tmpin[courant]->z ) / (float) ( tmpin[next]->z - tmpin[courant]->z );
           in[add]->x = ( tmpin[next]->x - tmpin[courant]->x ) * scale + tmpin[courant]->x;
           in[add]->y = ( tmpin[next]->y - tmpin[courant]->y ) * scale + tmpin[courant]->y;
           in[add]->z = ( tmpin[next]->z - tmpin[courant]->z ) * scale + tmpin[courant]->z;
           if (flags & INT_1COL)
              in[add]->c = ( tmpin[next]->c - tmpin[courant]->c ) * scale + tmpin[courant]->c;
           else
           {
              if (flags & INT_3COL)
              {
                 r = ((tmpin[next]->c & 0xFF0000) - (tmpin[courant]->c & 0xFF0000)) * scale + (tmpin[courant]->c & 0xFF0000);
                 g = ((tmpin[next]->c & 0x00FF00) - (tmpin[courant]->c & 0x00FF00)) * scale + (tmpin[courant]->c & 0x00FF00);
                 b = ((tmpin[next]->c & 0x0000FF) - (tmpin[courant]->c & 0x0000FF)) * scale + (tmpin[courant]->c & 0x0000FF);
                 in[add]->c =(r & 0xFF0000) | (g & 0x00FF00) | (b & 0x0000FF);
              }
              else
                 in[add]->c = tmpin[courant]->c;
           }
           if (flags & INT_UV)
           {
              in[add]->u = ( tmpin[next]->u - tmpin[courant]->u ) * scale + tmpin[courant]->u;
              in[add]->v = ( tmpin[next]->v - tmpin[courant]->v ) * scale + tmpin[courant]->v;
           }
           else
           {
             in[add]->u = tmpin[courant]->u;
             in[add]->v = tmpin[courant]->v;
           }

           // mise a jour des pteurs
           in[sortie] = in[add];

           // projection du nouveaux point
           persp_project_f( in[add]->x,in[add]->y,in[add]->z, &out[add]->x, &out[add]->y );
           out[add]->z = in[add]->z;
           out[add]->u = in[add]->u;
           out[add]->v = in[add]->v;
           out[add]->c = in[add]->c;

           // mise a jour pteur 2d
           out[sortie] = out[add];

           sortie++;
           add++;
        }
        // passage au segment suivant
        courant = next;
        test1   = test2;
     }


/*-----------------------------------------------------------------------------
 * test sur x < z
 *-----------------------------------------------------------------------------
 */
     for ( i = 0 ; i < sortie ; i++ ) tmpout[i] = out[i];
     for ( i = 0 ; i < sortie ; i++ ) tmpin[i]  = in[i];

     // test z < min_z
     vc      = sortie;
     courant = 0;
     sortie  = 0;

     test1   = (tmpin[courant]->x < tmpin[courant]->z);
     for ( i = 0 ; i < vc ; i++ )
     {
        next = (i == vc-1) ? 0 : i+1;

        // on garde le sommet courant s'il est visible
        if ( test1 )
        {
           out[sortie] = tmpout[courant];
           in[sortie]  = tmpin[courant];
           sortie++;
        }

        test2 = ( tmpin[next]->x < tmpin[next]->z );

        // ajouter un sommet couper si une bordure est dans l'ecran et l'autre hors
        if ( test1 != test2 )
        {
           // calcul des nouvelles valeurs
           scale = ( tmpin[courant]->z - tmpin[courant]->x ) /
                   ( tmpin[next]->x - tmpin[courant]->x - tmpin[next]->z + tmpin[courant]->z );
           in[add]->x = ( tmpin[next]->x - tmpin[courant]->x ) * scale + tmpin[courant]->x;
           in[add]->y = ( tmpin[next]->y - tmpin[courant]->y ) * scale + tmpin[courant]->y;
           in[add]->z = ( tmpin[next]->z - tmpin[courant]->z ) * scale + tmpin[courant]->z;
           if (flags & INT_1COL)
              in[add]->c = ( tmpin[next]->c - tmpin[courant]->c ) * scale + tmpin[courant]->c;
           else
           {
              if (flags & INT_3COL)
              {
                 r = ((tmpin[next]->c & 0xFF0000) - (tmpin[courant]->c & 0xFF0000)) * scale + (tmpin[courant]->c & 0xFF0000);
                 g = ((tmpin[next]->c & 0x00FF00) - (tmpin[courant]->c & 0x00FF00)) * scale + (tmpin[courant]->c & 0x00FF00);
                 b = ((tmpin[next]->c & 0x0000FF) - (tmpin[courant]->c & 0x0000FF)) * scale + (tmpin[courant]->c & 0x0000FF);
                 in[add]->c =(r & 0xFF0000) | (g & 0x00FF00) | (b & 0x0000FF);
              }
              else
                 in[add]->c = tmpin[courant]->c;
           }
           if (flags & INT_UV)
           {
              in[add]->u = ( tmpin[next]->u - tmpin[courant]->u ) * scale + tmpin[courant]->u;
              in[add]->v = ( tmpin[next]->v - tmpin[courant]->v ) * scale + tmpin[courant]->v;
           }
           else
           {
             in[add]->u = tmpin[courant]->u;
             in[add]->v = tmpin[courant]->v;
           }

           // mise a jour des pteurs
           in[sortie] = in[add];

           // projection du nouveaux point
           persp_project_f( in[add]->x,in[add]->y,in[add]->z, &out[add]->x, &out[add]->y );
           out[add]->z = in[add]->z;
           out[add]->u = in[add]->u;
           out[add]->v = in[add]->v;
           out[add]->c = in[add]->c;

           // mise a jour pteur 2d
           out[sortie] = out[add];

           sortie++;
           add++;
        }
        // passage au segment suivant
        courant = next;
        test1   = test2;
     }

/*-----------------------------------------------------------------------------
 * test sur x > -z
 *-----------------------------------------------------------------------------
 */
     for ( i = 0 ; i < sortie ; i++ ) tmpout[i] = out[i];
     for ( i = 0 ; i < sortie ; i++ ) tmpin[i]  = in[i];

     // test z < min_z
     vc      = sortie;
     courant = 0;
     sortie  = 0;

     test1   = (tmpin[courant]->x > -tmpin[courant]->z);
     for ( i = 0 ; i < vc ; i++ )
     {
        next = (i == vc-1) ? 0 : i+1;

        // on garde le sommet courant s'il est visible
        if ( test1 )
        {
           out[sortie] = tmpout[courant];
           in[sortie]  = tmpin[courant];
           sortie++;
        }

        test2 = (tmpin[next]->x > -tmpin[next]->z );

        // ajouter un sommet couper si une bordure est dans l'ecran et l'autre hors
        if ( test1 != test2 )
        {
           // calcul des nouvelles valeurs
           scale = (-tmpin[courant]->z - tmpin[courant]->x ) /
                   ( tmpin[next]->x - tmpin[courant]->x + tmpin[next]->z - tmpin[courant]->z);
           in[add]->x = ( tmpin[next]->x - tmpin[courant]->x ) * scale + tmpin[courant]->x;
           in[add]->y = ( tmpin[next]->y - tmpin[courant]->y ) * scale + tmpin[courant]->y;
           in[add]->z = ( tmpin[next]->z - tmpin[courant]->z ) * scale + tmpin[courant]->z;
           if (flags & INT_1COL)
              in[add]->c = ( tmpin[next]->c - tmpin[courant]->c ) * scale + tmpin[courant]->c;
           else
           {
              if (flags & INT_3COL)
              {
                 r = ((tmpin[next]->c & 0xFF0000) - (tmpin[courant]->c & 0xFF0000)) * scale + (tmpin[courant]->c & 0xFF0000);
                 g = ((tmpin[next]->c & 0x00FF00) - (tmpin[courant]->c & 0x00FF00)) * scale + (tmpin[courant]->c & 0x00FF00);
                 b = ((tmpin[next]->c & 0x0000FF) - (tmpin[courant]->c & 0x0000FF)) * scale + (tmpin[courant]->c & 0x0000FF);
                 in[add]->c =(r & 0xFF0000) | (g & 0x00FF00) | (b & 0x0000FF);
              }
              else
                 in[add]->c = tmpin[courant]->c;
           }
           if (flags & INT_UV)
           {
              in[add]->u = ( tmpin[next]->u - tmpin[courant]->u ) * scale + tmpin[courant]->u;
              in[add]->v = ( tmpin[next]->v - tmpin[courant]->v ) * scale + tmpin[courant]->v;
           }
           else
           {
             in[add]->u = tmpin[courant]->u;
             in[add]->v = tmpin[courant]->v;
           }

           // mise a jour des pteurs
           in[sortie] = in[add];

           // projection du nouveaux point
           persp_project_f( in[add]->x,in[add]->y,in[add]->z, &out[add]->x, &out[add]->y );
           out[add]->z = in[add]->z;
           out[add]->u = in[add]->u;
           out[add]->v = in[add]->v;
           out[add]->c = in[add]->c;

           // mise a jour pteur 2d
           out[sortie] = out[add];

           sortie++;
           add++;
        }
        // passage au segment suivant
        courant = next;
        test1   = test2;
     }

/*-----------------------------------------------------------------------------
 * test sur y < z
 *-----------------------------------------------------------------------------
 */
     for ( i = 0 ; i < sortie ; i++ ) tmpout[i] = out[i];
     for ( i = 0 ; i < sortie ; i++ ) tmpin[i]  = in[i];

     // test z < min_z
     vc      = sortie;
     courant = 0;
     sortie  = 0;

     test1   = (tmpin[courant]->y < tmpin[courant]->z);
     for ( i = 0 ; i < vc ; i++ )
     {
        next = (i == vc-1) ? 0 : i+1;

        // on garde le sommet courant s'il est visible
        if ( test1 )
        {
           out[sortie] = tmpout[courant];
           in[sortie]  = tmpin[courant];
           sortie++;
        }

        test2 = ( tmpin[next]->y < tmpin[next]->z );

        // ajouter un sommet couper si une bordure est dans l'ecran et l'autre hors
        if ( test1 != test2 )
        {
           // calcul des nouvelles valeurs
           scale = ( tmpin[courant]->z - tmpin[courant]->y ) /
                   ( tmpin[next]->y - tmpin[courant]->y - tmpin[next]->z + tmpin[courant]->z );
           in[add]->x = ( tmpin[next]->x - tmpin[courant]->x ) * scale + tmpin[courant]->x;
           in[add]->y = ( tmpin[next]->y - tmpin[courant]->y ) * scale + tmpin[courant]->y;
           in[add]->z = ( tmpin[next]->z - tmpin[courant]->z ) * scale + tmpin[courant]->z;
           if (flags & INT_1COL)
              in[add]->c = ( tmpin[next]->c - tmpin[courant]->c ) * scale + tmpin[courant]->c;
           else
           {
              if (flags & INT_3COL)
              {
                 r = ((tmpin[next]->c & 0xFF0000) - (tmpin[courant]->c & 0xFF0000)) * scale + (tmpin[courant]->c & 0xFF0000);
                 g = ((tmpin[next]->c & 0x00FF00) - (tmpin[courant]->c & 0x00FF00)) * scale + (tmpin[courant]->c & 0x00FF00);
                 b = ((tmpin[next]->c & 0x0000FF) - (tmpin[courant]->c & 0x0000FF)) * scale + (tmpin[courant]->c & 0x0000FF);
                 in[add]->c =(r & 0xFF0000) | (g & 0x00FF00) | (b & 0x0000FF);
              }
              else
                 in[add]->c = tmpin[courant]->c;
           }
           if (flags & INT_UV)
           {
              in[add]->u = ( tmpin[next]->u - tmpin[courant]->u ) * scale + tmpin[courant]->u;
              in[add]->v = ( tmpin[next]->v - tmpin[courant]->v ) * scale + tmpin[courant]->v;
           }
           else
           {
             in[add]->u = tmpin[courant]->u;
             in[add]->v = tmpin[courant]->v;
           }

           // mise a jour des pteurs
           in[sortie] = in[add];

           // projection du nouveaux point
           persp_project_f( in[add]->x,in[add]->y,in[add]->z, &out[add]->x, &out[add]->y );
           out[add]->z = in[add]->z;
           out[add]->u = in[add]->u;
           out[add]->v = in[add]->v;
           out[add]->c = in[add]->c;

           // mise a jour pteur 2d
           out[sortie] = out[add];

           sortie++;
           add++;
        }
        // passage au segment suivant
        courant = next;
        test1   = test2;
     }


/*-----------------------------------------------------------------------------
 * test sur y > -z
 *-----------------------------------------------------------------------------
 */
     for ( i = 0 ; i < sortie ; i++ ) tmpout[i] = out[i];
     for ( i = 0 ; i < sortie ; i++ ) tmpin[i]  = in[i];

     // test z < min_z
     vc      = sortie;
     courant = 0;
     sortie  = 0;

     test1   = (tmpin[courant]->y > -tmpin[courant]->z);
     for ( i = 0 ; i < vc ; i++ )
     {
        next = (i == vc-1) ? 0 : i+1;

        // on garde le sommet courant s'il est visible
        if ( test1 )
        {
           out[sortie] = tmpout[courant];
           in[sortie]  = tmpin[courant];
           sortie++;
        }

        test2 = ( tmpin[next]->y > -tmpin[next]->z);

        // ajouter un sommet couper si une bordure est dans l'ecran et l'autre hors
        if ( test1 != test2 )
        {
           // calcul des nouvelles valeurs
           scale = ( -tmpin[courant]->z - tmpin[courant]->y ) /
                   ( tmpin[next]->y - tmpin[courant]->y + tmpin[next]->z - tmpin[courant]->z );
           in[add]->x = ( tmpin[next]->x - tmpin[courant]->x ) * scale + tmpin[courant]->x;
           in[add]->y = ( tmpin[next]->y - tmpin[courant]->y ) * scale + tmpin[courant]->y;
           in[add]->z = ( tmpin[next]->z - tmpin[courant]->z ) * scale + tmpin[courant]->z;
           if (flags & INT_1COL)
              in[add]->c = ( tmpin[next]->c - tmpin[courant]->c ) * scale + tmpin[courant]->c;
           else
           {
              if (flags & INT_3COL)
              {
                 r = ((tmpin[next]->c & 0xFF0000) - (tmpin[courant]->c & 0xFF0000)) * scale + (tmpin[courant]->c & 0xFF0000);
                 g = ((tmpin[next]->c & 0x00FF00) - (tmpin[courant]->c & 0x00FF00)) * scale + (tmpin[courant]->c & 0x00FF00);
                 b = ((tmpin[next]->c & 0x0000FF) - (tmpin[courant]->c & 0x0000FF)) * scale + (tmpin[courant]->c & 0x0000FF);
                 in[add]->c =(r & 0xFF0000) | (g & 0x00FF00) | (b & 0x0000FF);
              }
              else
                 in[add]->c = tmpin[courant]->c;
           }
           if (flags & INT_UV)
           {
              in[add]->u = ( tmpin[next]->u - tmpin[courant]->u ) * scale + tmpin[courant]->u;
              in[add]->v = ( tmpin[next]->v - tmpin[courant]->v ) * scale + tmpin[courant]->v;
           }
           else
           {
             in[add]->u = tmpin[courant]->u;
             in[add]->v = tmpin[courant]->v;
           }

           // mise a jour des pteurs
           in[sortie] = in[add];

           // projection du nouveaux point
           persp_project_f( in[add]->x,in[add]->y,in[add]->z, &out[add]->x, &out[add]->y );
           out[add]->z = in[add]->z;
           out[add]->u = in[add]->u;
           out[add]->v = in[add]->v;
           out[add]->c = in[add]->c;

           // mise a jour pteur 2d
           out[sortie] = out[add];

           sortie++;
           add++;
        }
        // passage au segment suivant
        courant = next;
        test1   = test2;
     }

     // nombre de vertex ajoutes
     free (tmpin);
     free (tmpout);
     if ( sortie < 3 )
     {
        *nbad = 0;
        return 0;
     }
     else
     {
       if (type == POLYTYPE_FLAT)
          out[0]->c = color;
       *nbad = add - base;
       return (sortie);
     }
 }


 #undef INT_NONE
 #undef INT_1COL
 #undef INT_3COL
 #undef INT_UV









