/*================================================
 * lib de chargement d'objet LightWave
 *------------------------------------------------
 * (c) Disk91 / TeKnO-BeDoUiNs / 1998
 *================================================
 */

 #include <stdlib.h>
 #include <stdio.h>
 #include <allegro.h>
 #include "3d.h"
 #include "lwo.h"

 int lwo_vtx_by_poly;


 /*--------------------------------------------
  * Trouve le nombre de points et faces
  * d'un objet et le nombre max de vtx par face
  *--------------------------------------------
  * revoie -1 si erreur
  *--------------------------------------------
  */
  static int get_infos( int * nbvtx, int * nbpoly, int * max_vtx, char * filename )
  {
     PACKFILE * file;

     char     * token;
     int        i;
     unsigned char  * buffer;
     int              bufl;
     int              nbvtxfac;         // nombre de vtx dans la face courante

     file = pack_fopen( filename, "rb" );
     if ( file )
     {
        token  = lwo_get_id4( file );
        if ( strcmp( token, "FORM" ) )   /*--- test for LW file ---*/
           LWOERROR("Fichier non standard LWO",51)
        lwo_get_int( file );                 /*--- jump for data size ---*/

        token  = lwo_get_id4( file );
        if ( strcmp( token, "LWOB" ) )   /*--- test for LWO file ---*/
           LWOERROR("Fichier non standard LWO",51)

       /*--- get the vertex (12 byte per vertex) ---*/
       if ( !lwo_find_chunck( file,"PNTS" ) )      /*--- find points 'PNTS' chunck ---*/
           LWOERROR("Fichier non standard LWO",51)

       *nbvtx = lwo_get_int( file ) / 12;
       for ( i = 0 ; i < *nbvtx * 3 ; i++ )
          lwo_get_float( file );


       /*--- get the polygons ---*/
       if ( !lwo_find_chunck( file,"POLS" ) )        /*--- find poly 'POLS' chunck ---*/
           LWOERROR("Fichier non standard LWO",51)

       bufl = lwo_get_int( file );                  /*--- create a buffer for scaning ---*/
       buffer = (char *) malloc ( bufl );
       pack_fread( buffer, bufl, file);
       *nbpoly  = 0;
       *max_vtx = 0;                            /*--- nombre max de vertex par face ---*/
       for ( i = 0 ; i < bufl ; i+= ( 4 + nbvtxfac * 2 ) )
       {
          nbvtxfac = (buffer[i+1]+(buffer[i] << 8));
          if ( nbvtxfac > *max_vtx )            /*--- nb max de vtx dans les faces ---*/
             *max_vtx = nbvtxfac;

          (*nbpoly)++;                            /*--- number of polygons ---*/
       }

       free(buffer);
       pack_fclose(file);
       return ( 0 );
    }
    return ( -1 );
 }



 /*=====================================================================================
  * public for OBJECT_f
  *=====================================================================================
  */
  OBJECT_f * load_lwo_f( char * filename )
  {
    PACKFILE * file;
    OBJECT_f * object;
    int        nbvtx;                   // nombre de point dans l'objet
    int        nbpoly;                  // nombre de polygones dans l'objet
    int        maxvtx;                  // nombre max de vtx dans les polygones

    int        i,j,k;

    if ( get_infos( &nbvtx, &nbpoly, &maxvtx, filename ) != -1 )
    {
       lwo_vtx_by_poly = maxvtx;
       file = pack_fopen( filename, "rb" );

       // creation de l'objet avec les infos de get_infos
       object = create_object_f ( nbvtx, nbpoly, maxvtx );

       lwo_get_id4( file );                         // jump FROM chunck
       lwo_get_int( file );                         // jump data size
       lwo_get_id4( file );                         // jump LWO chunck

       // reading vertex informations
       lwo_find_chunck( file,"PNTS" );              // find PNTS chunck
       lwo_get_int( file );                         // jump vtx size
       for ( i = 0 ; i < object->nb_vtx ; i++ )
       {
          object->vtx_begin[i].x =   lwo_get_float( file );
          object->vtx_begin[i].y =   lwo_get_float( file );
          object->vtx_begin[i].z =   lwo_get_float( file );
          object->vtx_begin[i].c =   1;
       }

       // reading polygons informations
       lwo_find_chunck( file,"POLS" );              // find POLS chunck
       lwo_get_int( file );                         // jump poly size
       for ( k=0 , i = 0 ; i < object->nb_poly_begin ; i++)
       {
          // for each polygons reading vertex informations
          object->poly_begin[i].nb_vtx = (int) lwo_get_sint( file );
          for ( j = 0 ; j < object->poly_begin[i].nb_vtx ; j++ )
          {
             object->poly_begin[i].index[j] = lwo_get_sint( file );
             object->poly_begin[i].ptvtx3d[j] = &object->vtx_begin[object->poly_begin[i].index[j]];
             object->poly_begin[i].u[j] = 0.0;
             object->poly_begin[i].v[j] = 0.0;
          }

          // getting polygon sup informations ( polygon name )
          object->poly_begin[i].coul = lwo_get_sint( file );
       }

       pack_fclose(file);

       return ( object );
    }
    else
      LWOERROR("Impossible de lire le fichier .lwo",50)
 }



