/*=======================================================
 * Traitement des object 3D lwo pour les transformations
 *-------------------------------------------------------
 * (c) Disk91 / TeKnO-BeDoUiNs / 11-1998
 *=======================================================
 */
 #include <math.h>
 #include <stdio.h>
 #include <allegro.h>
 #include "3d.h"
 extern BITMAP * dblbuff;

   RGB     coul = { 63,45,18,0 };
   RGB     coul1 = { 63,0,0,0 };
   RGB     blk  = { 0,0,0,0 };

 #define MAXPILE   40000

/*---------------------------------------------
 * helper : donne le vecteur normal a une face
 * renvoie -1 si la face doit etre affichee par
 *  defaut de normale valable
 *---------------------------------------------
 */
 static int v_norm( V3D_f ** ptpoly, int nbpoly, float * vnx, float * vny, float * vnz )
 {
    V3D_f * v2, * v3;
    V3D_f va,vb;
    int  k,z;

    v2 = ptpoly[0];

    // objectif : trouver 2 vecteurs non nuls et non colineaires

    // trouver le premier E0: du point 0 vers un autre trouve dans sens aiguilles montre
    k = 0;
    do
    {
       if ( k < nbpoly ) k++; else return (-1);

       v3 = ptpoly[k];
       va.x = (v3->x - v2->x);
       va.y = (v3->y - v2->y);
       va.z = (v3->z - v2->z);
    } while ( va.x == 0 && va.y == 0 && va.z == 0 );

    // trouver le premier E1: du point 0 vers un autre trouve dans sens inv aiguilles montre
    z = nbpoly;
    do
    {
      do
      {
         if ( z > k )  z--; else return (-1);
         v3 = ptpoly[z];
         vb.x = (v3->x - v2->x);
         vb.y = (v3->y - v2->y);
         vb.z = (v3->z - v2->z);

      } while ( vb.x == 0 && vb.y == 0 && vb.z == 0 );
      cross_product_f(  va.x, va.y, va.z, vb.x, vb.y, vb.z, vnx, vny, vnz );
    } while ( *vnx == 0 && *vny == 0 && *vnz == 0 );
    return 0;
 }


/*-------------------------------------------
 * Calcul des normales aux faces pour un
 * objet => elimination faces cachees
 * les normales sont calculees dans l'espace
 * objet c'est l'orbservateur qui se deplace
 *-------------------------------------------
 */
 void normalize_object_f( OBJECT_f * ob  )
 {
     int        i;
     float      nx,ny,nz;                     // stocke les normales temporairement

     for ( i = 0 ; i < ob->nb_poly_begin ; i++ )
     {
         if ( ob->poly_begin[i].nb_vtx > 2 )
         {
            // calcul le vect normal
            if ( !v_norm( ob->poly_begin[i].ptvtx3d,ob->poly_begin[i].nb_vtx,&nx,&ny,&nz ) )
            {
               normalize_vector_f(&nx,&ny,&nz);
               ob->poly_begin[i].nx = nx;
               ob->poly_begin[i].ny = ny;
               ob->poly_begin[i].nz = nz;
               ob->poly_begin[i].state = 1;
            }
            else        // impossible de determiner un vercteur utilisable
            {
               ob->poly_begin[i].nx = 0.0;
               ob->poly_begin[i].ny = 0.0;
               ob->poly_begin[i].nz = 0.0;
               ob->poly_begin[i].state = 0;
            }
        }
        else   // trop peu de points pour avoir un vecteur utilisable
        {
           ob->poly_begin[i].nx = 0.0;
           ob->poly_begin[i].ny = 0.0;
           ob->poly_begin[i].nz = 0.0;
           ob->poly_begin[i].state = 0;
        }
    }
}



/*-------------------------------------------
 * Transformation d'un objet
 *-------------------------------------------
 * Entree : objet et donnee des transfo
 * Sortie : une nouvel objet cree
 *-------------------------------------------
 */
 void transform_object_f( OBJECT_f * ob, float scale, float xrot, float yrot, float zrot,
                                                      float tx,   float ty,   float tz   )
 {
     MATRIX_f   m;                            // pour la transfo des objets
     MATRIX_f   m1,m2,m3;                     // pour la transfo de l'observateur
     int        i,k,j;

     V3D_f   *  ptvtx;                        // des pteur temporaire sur des faces
     V3D_f   *  ptvtx1;
     V3D_f   *  ptvtx2;

     V3D_f      observ;                       // observateur
     float      ox,oy,oz;                     // vecteur obs->point
     int     *  pts;                          // liste des points visibles

     POLY_f  *  poly;                         // pteur poly pour simplifier ecritures
     POLY_f  *  poly2;

     // init
     pts = ( int * ) malloc ( ob->nb_vtx * sizeof( int ) );
     if ( !pts ) OBJECTERROR("Impossible d'allouer la memoire au buffer vertex",80)

     for ( i = 0 ; i < ob->nb_vtx ; i++ ) pts[i] = 0;


     // applique la transformation sur l'observatur
     //  remarque, la rotation est differente parcequ'il faut suivre le chemin
     //  inverse de conception de la matrice : multiplication de matrice
     //  est non commutative ...
     observ.x = -tx;
     observ.y = -ty;
     if ( scale != 0.0 )
        observ.z = -tz/(float)scale;
     get_transformation_matrix_f(&m1, 1.0, 255.0-xrot, 0.0,  0.0, 0.0, 0.0, 0.0);
     get_transformation_matrix_f(&m2, 1.0, 0.0, 255.0-yrot,  0.0, 0.0, 0.0, 0.0);
     get_transformation_matrix_f(&m3, 1.0, 0.0, 0.0, 255.0-zrot,  0.0, 0.0, 0.0);
     apply_matrix_f( &m1, observ.x,observ.y,observ.z,&observ.x,&observ.y,&observ.z );
     apply_matrix_f( &m2, observ.x,observ.y,observ.z,&observ.x,&observ.y,&observ.z );
     apply_matrix_f( &m3, observ.x,observ.y,observ.z,&observ.x,&observ.y,&observ.z );


     // determiner les faces visibles
     for ( i = 0 ; i < ob->nb_poly_begin ; i++ )
     {
        poly = &ob->poly_begin[i];
        // verifie que l'on ait un vecteur normal utilisable pour cette face
        if ( poly->state )
        {
          // calcul le vecteur observateur --> point
          ox = poly->ptvtx3d[0]->x - observ.x;
          oy = poly->ptvtx3d[0]->y - observ.y;
          oz = poly->ptvtx3d[0]->z - observ.z;
          // calcul du produit scalaire => angle
          if ( dot_product_f(poly->nx, poly->ny, poly->nz, ox, oy, oz) <= 0.0 )
          {
              // ajoute les points visible a la liste
              for ( j = 0 ; j < poly->nb_vtx ; j++ )
                  pts[ poly->index[j] ] = 1;
              // ajoute la faces a la liste des visibles
              ob->poly_end[i].state = 1;
          }
          else
             ob->poly_end[i].state = 0;
        }
        else
        {
           // ajoute les points visible a la liste
           for ( j = 0 ; j < poly->nb_vtx ; j++ )
               pts[ poly->index[j] ] = 1;
           // ajoute la faces a la liste des visibles
           ob->poly_end[i].state = 1;
        }
     }


     // recopie des faces visibles
     for ( k = 0,i = 0 ; i < ob->nb_poly_begin ; i++ )
     {
         poly  = &ob->poly_begin[i];
         poly2 = &ob->poly_end[k];
         if (  ob->poly_end[i].state )
         {
            poly2->nb_vtx = poly->nb_vtx;
            poly2->coul   = poly->coul;
            for ( j = 0 ; j < poly->nb_vtx ; j++ )
            {
              poly2->ptvtx3d[j] = &ob->vtx_end[ poly->index[j] ];
              poly2->ptvtx2d[j] = &ob->vtx_2d [ poly->index[j] ];
            }
            k++;
         }
     }
     ob->nb_poly_end = k;


     // transforme les points de l'objet de base s'ils sont visibles et projette
     get_transformation_matrix_f(&m, scale, xrot, yrot, zrot, tx, ty, tz);
     for ( i = 0 ; i < ob->nb_vtx ; i++ )
     {
        if ( pts[i] == 1 )
        {
           ptvtx  = &ob->vtx_begin[i];
           ptvtx1 = &ob->vtx_end[i];
           ptvtx2 = &ob->vtx_2d[i];

           // transformation
           apply_matrix_f( &m, ptvtx->x, ptvtx->y, ptvtx->z, &ptvtx1->x, &ptvtx1->y, &ptvtx1->z);
           ptvtx1->c = ptvtx->c;

           // projection
           persp_project_f( ptvtx1->x,ptvtx1->y,ptvtx1->z, &ptvtx2->x, &ptvtx2->y );
           ptvtx2->c = ptvtx->c;
           ptvtx2->z = ptvtx1->z;

        }
     }
     free (pts);
 }

/*-------------------------------------------
 * Transformation d'un objet issue d'un bsp
 *-------------------------------------------
 * Entree : objet et donnee des transfo
 * Sortie : une nouvel objet cree
 *-------------------------------------------
 */
 void transform4bsp_object_f( OBJECT_f * ob, float scale, float xrot, float yrot, float zrot,
                                                      float tx,   float ty,   float tz   )
 {
     MATRIX_f   m;                            // pour la transfo des objets
     int        i,j;

     V3D_f   *  ptvtx;                        // des pteur temporaire sur des faces
     V3D_f   *  ptvtx1;
     V3D_f   *  ptvtx2;

     int        pts[MAXPILE];                 // liste des points visibles

     POLY_f  *  poly;                         // pteur poly pour simplifier ecritures

     for ( i = 0 ; i < ob->nb_vtx ; i++ ) pts[i] = 0;

     // creer la liste des points utilises pour l'objet
     for ( i = 0 ; i < ob->nb_poly_end ; i++ )
     {
        poly = &ob->poly_end[i];
        // ajoute les points visible a la liste
        for ( j = 0 ; j < poly->nb_vtx ; j++ )
            pts[ poly->index[j] ] = 1;
     }

     // transforme les points de l'objet de base s'ils sont visibles et projette
     get_transformation_matrix_f(&m, scale, xrot, yrot, zrot, tx, ty, tz);
     for ( i = 0 ; i < ob->nb_vtx ; i++ )
     {
        if ( pts[i] == 1 )
        {
           ptvtx  = &ob->vtx_begin[i];
           ptvtx1 = &ob->vtx_end[i];
           ptvtx2 = &ob->vtx_2d[i];

           // transformation
           apply_matrix_f( &m, ptvtx->x, ptvtx->y, ptvtx->z, &ptvtx1->x, &ptvtx1->y, &ptvtx1->z);
           ptvtx1->c = ptvtx->c;

           // projection
           persp_project_f( ptvtx1->x,ptvtx1->y,ptvtx1->z, &ptvtx2->x, &ptvtx2->y );
           ptvtx2->c = ptvtx->c;
           ptvtx2->z = ptvtx1->z;

        }
     }
 }



/*-------------------------------------------
 * tri_fnc : trie sur z fonction utilisee par
 *           q sort
 *-------------------------------------------
 */
 int tri_fnc( const void * c1, const void * c2 )
 {
    POLY_f *  f1, *  f2;
    f1 = ( POLY_f * ) *(( POLY_f ** )c1);
    f2 = ( POLY_f * ) *(( POLY_f ** )c2);

    if ( f2->zmoy > f1->zmoy ) return  1;
    if ( f2->zmoy < f1->zmoy ) return -1;
    return 0;
 }


/*-------------------------------------------
 * Projection clip : realise le clipping de
 *  face et la projection 2D
 *-------------------------------------------
 * Entree : Objet + info de cliping
 *          en entree est passe un objet2d,
 *          si il est null, il est cree, sinon
 *          on le reutilise : plein de tps
 *          gagne...
 * Sortie : OBJECT2D_f en float
 *-------------------------------------------
 */
 void clip3d_object_f( OBJECT_f * ob, int type, float min_z, float max_z )
 {

    float   xmax,xmin, ymax, ymin, zmax, zmin;
    float   abszmax,abszmin;
    float   tmpz,tmpz1;                  // tmp pour abs zmax

    int     i,j,k;

    POLY_f  * poly;                     // pteur sur polygon temporaire

    int     ajout;                      // index dans les points ajoutes
    int     nbad;                       // nombre de vtx ajoutes lors d'un clip

    // init de ajout : index sur points libre
    ajout = ob->nb_vtx;

    // Clipping des faces en 3d pour eliminer celle hors champ
    for ( k = 0, i = 0 ; i < ob->nb_poly_end ; i++ )
    {
       poly  = &ob->poly_end[i];

       //--- recherche des maximums et minimum de la face ---
       xmin         = poly->ptvtx3d[0]->x;
       xmax         = xmin;
       ymin         = poly->ptvtx3d[0]->y;
       ymax         = ymin;
       zmin         = poly->ptvtx3d[0]->z;
       zmax         = zmin;
       for ( j = 1 ; j < poly->nb_vtx ; j++ )
       {
         if ( poly->ptvtx3d[j]->x > xmax )
            xmax = poly->ptvtx3d[j]->x;
         else
            if ( poly->ptvtx3d[j]->x < xmin )
               xmin = poly->ptvtx3d[j]->x;

         if ( poly->ptvtx3d[j]->y > ymax )
            ymax = poly->ptvtx3d[j]->y;
         else
            if ( poly->ptvtx3d[j]->y < ymin )
               ymin = poly->ptvtx3d[j]->y;

         if ( poly->ptvtx3d[j]->z > zmax )
            zmax = poly->ptvtx3d[j]->z;
         else
            if ( poly->ptvtx3d[j]->z < zmin )
               zmin = poly->ptvtx3d[j]->z;
       }

       //--- exclusion des polygons sur z_min et z_max ---
       if ( zmax <= min_z || zmin >= max_z )
          continue;

       //--- exclusion des polygons sur x et y hors champ ---
       abszmax = (float)fabs((double)zmax);
       abszmin = (float)fabs((double)zmin);
       if ( abszmax > abszmin )
       {
          tmpz  = abszmax;
          tmpz1 = abszmin;
       }
       else
       {
          tmpz1 = abszmax;
          tmpz  = abszmin;
       }

       if (
                 xmax < -tmpz || xmin > tmpz
             ||  ymax < -tmpz || ymin > tmpz

          )
          continue;

       //--- les faces incluses completement ---
       if (  zmax > min_z  &&
             xmax > -tmpz1 && xmin < tmpz1 &&
             ymax > -tmpz1 && ymin < tmpz1 &&
             xmin > -tmpz1 && xmax < tmpz1 &&
             ymin > -tmpz1 && ymax < tmpz1

          )
       {
          // recopie le infos --> ajoute le pteur de face
          poly->zmoy = (zmax + zmin) / 2.0;
          ob->ptpoly[k] = poly;
          k++;
       }
       //les faces dont on est pas sure
       else
       {
          // prepare des pteurs suplementaires pour le clipping
          // ici 10 car on a mis DISK_MAX = nb vtx max + 10 et j'espere que ca suffis tjrs
          for ( j = 0 ; j < 10 ; j++ )
          {
             poly->ptvtx3d[j+poly->nb_vtx] = &ob->vtx_end[ajout+j];
             poly->ptvtx2d[j+poly->nb_vtx] = &ob->vtx_2d[ajout+j];
          }

          // clipping sur z_min, le reste est fait par allegro et zmax ignore
          poly->nb_vtx = clip_f( type, min_z, &nbad, poly->nb_vtx, poly->ptvtx3d, poly->ptvtx2d );
          // on complete la structure
          if ( poly->nb_vtx > 0 )
          {
             ajout += nbad;
             poly->zmoy = (zmax + zmin) / 2.0;
             ob->ptpoly[k] = poly;
             k++;
          }
       }
    }
    ob->nb_poly_end = k;

    // trie des faces sur z
    qsort( ob->ptpoly, k, sizeof( POLY_f * ), tri_fnc );

 }


/*-------------------------------------------
 * Projection clip : realise le clipping de
 *  face et la projection 2D
 * pour arbre bsp
 *-------------------------------------------
 * Entree : Objet + info de cliping
 *          en entree est passe un objet2d,
 *          si il est null, il est cree, sinon
 *          on le reutilise : plein de tps
 *          gagne...
 * Sortie : OBJECT2D_f en float
 *-------------------------------------------
 */
 void clip3d4bsp_object_f( OBJECT_f * ob, int type, float min_z, float max_z )
 {


    int     i,j,k;

    POLY_f  * poly;                     // pteur sur polygon temporaire

    int     ajout;                      // index dans les points ajoutes
    int     nbad;                       // nombre de vtx ajoutes lors d'un clip

    // init de ajout : index sur points libre
    ajout = ob->nb_vtx;

    // Clipping des faces en 3d pour eliminer celle hors champ
    for ( k = 0, i = 0 ; i < ob->nb_poly_end ; i++ )
    {
       poly  = &ob->poly_end[i];

       // les faces incluses completement
       if ( poly->state == 0 )
       {
          ob->ptpoly[k] = poly;
          k++;
       }
       // les faces clippees
       else
       {
          // prepare des pteurs suplementaires pour le clipping
          // ici 10 car on a mis DISK_MAX = nb vtx max + 10 et j'espere que ca suffis tjrs
          for ( j = 0 ; j < 10 ; j++ )
          {
             poly->ptvtx3d[j+poly->nb_vtx] = &ob->vtx_end[ajout+j];
             poly->ptvtx2d[j+poly->nb_vtx] = &ob->vtx_2d[ajout+j];
          }

          // clipping sur z_min, le reste est fait par allegro et zmax ignore
          poly->nb_vtx = clip_f( type, min_z, &nbad, poly->nb_vtx, poly->ptvtx3d, poly->ptvtx2d );
          // on complete la structure
          if ( poly->nb_vtx > 2 )
          {
             ajout += nbad;
             ob->ptpoly[k] = poly;
             k++;
          }
       }
    }
    ob->nb_poly_end = k;
 }

#undef MAXPILE
